/*
 * $RCSfile: ethdev.c,v $
 * 
 * x-kernel v3.2
 *
 * Copyright (c) 1996  Arizona Board of Regents
 *
 * $Log: ethdev.c,v $
 * Revision 1.5  1997/04/23 19:06:26  mjk
 * Switched to using /dev/ltap
 *
 * Revision 1.4  1997/03/10 21:37:08  mjk
 * *** empty log message ***
 *
 * Revision 1.3  1996/11/22 18:42:56  mjk
 * *** empty log message ***
 *
 * Revision 1.1  1996/07/24 00:00:13  mjk
 * Initial revision
 *
 */


#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <signal.h>
#include <ctype.h>
#include <assert.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <net/if.h>
#include <linux/if_ether.h>
#include <linux/ethtap.h>
#include "xkernel.h"
#include "eth.h"
#include "eth_i.h"
#include "machine.h"
#include "inputProcess.h"
#include "romopt.h"
#include "thread.h"


/* 
 * Defines 
 */

#define ETHDEV_NUMTHREADS	16	/* # of xk event rx handlers 	*/
/* max rx packet size 		*/
#define ETHDEV_BUFSIZE		(MAX_ETH_DATA_SZ + 100)

#ifdef XK_DEBUG
/* #define XK_RX_DEBUG */
/* #define XK_TX_DEBUG */
#endif



/* 
 * Typedefs 
 */

typedef struct {		/* Protocol State 			*/
  ETHhost	ethaddr;		/* ethernet address		*/
  int		ltap;			/* /dev/ltap fd 		*/
  pthread_t	tid;			/* recieving thread 		*/
				        /* rx thread attributes 	*/
  pthread_attr_t	thread_attributes;
  BufferPool		*buffers;	/* incomming data buffer pool	*/
  int			shutdown;	/* shutdown predicate 		*/
} PState;



/* 
 * Prototypes 
 */

void			ethdev_init(XObj);
static xkern_return_t	ethdev_openenable(XObj, XObj, XObj, Part *);
static xkern_return_t	ethdev_shutdown(XObj);
static xmsg_handle_t	ethdev_push(XObj, Msg *);
static int		ethdev_control(XObj, int, char *, int);

static long		ethdev_msgload(void *, char *, long, void *);
static void		ethdev_internal_demux(InputBuffer *);

static void		*ethdev_rxthread(void *);
static void		ethdev_gethost(XObj);

#ifdef XK_DEBUG
static void		ethdev_dump(XObj, FILE *, const char *,
				    unsigned char *, int);
#endif
static xkern_return_t	ethdev_nonpromiscuous(XObj);


/* 
 * Globals 
 */

int			traceethdevp;	/* trace variable 		*/


					/* rom options			*/
static XObjRomOpt	ethdev_opts[] = {
};



/* 
 * Public Functions
 */

void
ethdev_init(XObj self) 
{
  PState		*ps;
#ifndef ETHDEV_NOWRITE
  int			mode = ETHTAP_ACTIVE_MODE;
#endif  


  xTraceP0(self, TR_GROSS_EVENTS, "init");
  ps = X_NEW(PState);
  ps->shutdown = 0;
  self->state = ps;

#if 0  
  				/* Parse ROM file 			*/
  findXObjRomOpts(self, ethdev_opts,
		  (sizeof ethdev_opts) / sizeof(XObjRomOpt), NULL);
#endif  
  

  if ( (ps->ltap = open("/dev/ltap", O_RDWR|O_NONBLOCK)) == -1 ) {
    xError("cannot open /dev/ltap");
    perror("");
    exit(-1);
  }

#ifdef ETHDEV_NOWRITE
  xTraceP0(self, TR_GROSS_EVENTS, "passive mode");
#else
  if ( ioctl(ps->ltap, ETHTAPSETMODE, &mode) ) {
    xError("xkout ioctl ETHTAPSETMODE failed");
    exit(-1);
  }
#endif
  
  ethdev_gethost(self);	/* read eth0 address 				*/
#if 0  
  ethdev_nonpromiscuous(self);	/* non-promiscuous mode 		*/
#endif  

  ps->buffers = xkBufferPoolInit(ETHDEV_NUMTHREADS, ETHDEV_BUFSIZE,
				 ethdev_internal_demux);


  self->openenable	= ethdev_openenable;
  self->shutdown	= ethdev_shutdown;
  self->control		= ethdev_control;
  self->push		= ethdev_push;
  self->up		= NULL;
  

} /* ethdev_init */



/* 
 * Public Protocol Methods 
 */


static xkern_return_t
ethdev_openenable(XObj self, XObj hlpRcv, XObj hlpType, Part *p)
{
  struct sched_param	sched;
  PState		*ps = self->state;

  if ( self->up ) {
    xError("ethdev_openenable called multiple times!");
    return XK_FAILURE;
  }
  
  self->up = hlpRcv;

				/* Start rxer thread			*/
  pthread_attr_init(&ps->thread_attributes);
  pthread_attr_getschedparam(&ps->thread_attributes, &sched);
  sched.sched_priority = THREAD_STDPRIO - 10;	/* NOT POSIX		*/
  pthread_attr_setschedparam(&ps->thread_attributes, &sched);
  pthread_create(&ps->tid, &ps->thread_attributes, ethdev_rxthread, self);
  
  return XK_SUCCESS;
} /* ethpth_openable */


static xkern_return_t
ethdev_shutdown(XObj self) 
{
  PState		*ps = self->state;
 
  xTraceP0(self, TR_MAJOR_EVENTS, "shutdown");
  ps->shutdown = 1;		/* shutdown driver threads 		*/
  close(ps->ltap);

  return XK_SUCCESS;
} /* ethdev_shutdown */


static int
ethdev_control(XObj self, int opcode, char *buf, int buflen)
{
  PState	*ps = (PState *)self->state;
   
  switch ( opcode ) {

    case GETMYHOST:
      xTraceP0(self, TR_MAJOR_EVENTS, "getmyhost");
      checkLen(buflen, sizeof(ETHhost));
      memcpy(buf, &ps->ethaddr, sizeof(ETHhost));
      return sizeof(ETHhost);

    case ETH_SETPROMISCUOUS:
      return -1;		/* do not allow promiscous mode 	*/

    default:
      return -1;
  }
} /* ethdev_control */


static xmsg_handle_t
ethdev_push(XObj self, Msg *packet)
{
  PState		*ps 	= (PState *)self->state;
  ETHhdr		*header	= msgGetAttr(packet, 0);
  unsigned char		buf[ETHDEV_BUFSIZE];
  unsigned char		*numpkts = buf;
  unsigned short	*pktlen  = (unsigned short *)&buf[1];
  unsigned char		*data	 = (unsigned char *)&buf[3];
  int			sentlen;

  xTraceP0(self, TR_EVENTS, "push");

  *numpkts = 1;			/* write one packet at a time 		*/
  
  if ( !msgGetAttr(packet, 0) ) {
    xTraceP0(self, TR_ERRORS, "push -- no header attached");
    return XMSG_ERR_HANDLE;
  }

  if ( (*pktlen = msgLen(packet)) > ETHDEV_BUFSIZE ) {
    xTraceP1(self, TR_ERRORS, "push -- msgLen(%d) too big", msgLen(packet));
    return XMSG_ERR_HANDLE;
  }

				/* Create a flat buffer with the
				   complete ethernet packet */
  memcpy(data, header, sizeof(ETHhdr));

				/* Don't send broadcast packet that
				   did not originate from this
				   host. */
  if ( memcmp(&header->src, &ps->ethaddr, sizeof(ETHhost)) )
    return XMSG_NULL_HANDLE;

  xTraceP4(self, TR_MAJOR_EVENTS, "push: %04x %s > %s %d bytes",
	   ntohs(header->type),
	   ethHostStr(&header->src), ethHostStr(&header->dst),
	   msgLen(packet));

  msg2buf(packet, &data[sizeof(ETHhdr)]);
  *pktlen += sizeof(ETHhdr);	/* add header to length 		*/
  if ( *pktlen < ETH_ZLEN )	/* pad to minimum length (w/o FCS)	*/
    *pktlen = ETH_ZLEN;

#ifdef XK_TX_DEBUG
  xIfTraceP(self, TR_EVENTS) {
    ethdev_dump(self, stdout, "TX", numpkts,
		*pktlen + sizeof(char) + sizeof(short));
  }
#endif  

  sentlen = write(ps->ltap, numpkts, *pktlen + sizeof(char) + sizeof(short));

  if ( sentlen <= 0 ) {
    xTraceP0(self, TR_ERRORS, "push: write failed");
    return XMSG_ERR_HANDLE;
  }

  xTraceP1(self, TR_MORE_EVENTS, "msg sent %d bytes", sentlen);

  return XMSG_NULL_HANDLE;
} /* ethdev_push */



/* 
 * Private Protocol Methods 
 */

static void
ethdev_internal_demux(InputBuffer *buffer)
{
  ETHhdr	header;
  Msg		*msg	= &buffer->msg;
  XObj		self	= buffer->self;
  
  if ( !msgPop(msg, ethdev_msgload, &header, sizeof(ETHhdr), NULL) ) {
    xTraceP0(self, TR_SOFT_ERRORS,
	     "demux: incoming message too small!");
    msgDestroy(msg);
    return;
  }
  if ( !self->up ) {
    xTraceP0(self, TR_ERRORS, "demux: no upper protocol!");
    msgDestroy(msg);
    return;
  }

  /* dump packet header			*/
  xTraceP4(self, TR_MAJOR_EVENTS, "demux: %04x %s > %s %d bytes",
	   ntohs(header.type),
	   ethHostStr(&header.src), ethHostStr(&header.dst),
	   msgLen(msg));

  msgSetAttr(msg, 0, &header, sizeof(ETHhdr));

  xDemux(self, msg);		/* send to eth protocol 		*/
} /* ethdev_internal_demux */


static long
ethdev_msgload(void *header, char *buf, long buflen, void *arg )
{
  xAssert(buflen == sizeof(ETHhdr));
  memcpy(header, buf, sizeof(ETHhdr));
  return sizeof(ETHhdr);
} /* ethdev_msgload */



/* 
 * Private non-xKernel driver method.
 */

				/* The following routines are called
                                   outside of xKernel control so
                                   careful use of xk_master_lock() and
                                   xk_master_unlock() is required. */

static void *
ethdev_rxthread(void *arg)
{
  XObj			self    = (XObj)arg;
  PState		*ps	= (PState *)self->state;

  xk_master_lock();		/* Lock xKernel				*/
  
  while ( ! ps->shutdown ) {
    unsigned char	buf[0x4000];
    unsigned char	*p 	= buf;
    unsigned char	pkts;
    unsigned short	len;
    InputBuffer		*buffer;
    int			i;

    if ( read(ps->ltap, buf, sizeof buf) < 0 ) {
      xError("ethdev ERROR: Cannot read from ethernet");
      continue;
    }

    pkts = *p++;
    if ( pkts ) {
      xTraceP1(self, TR_EVENTS, "read %d packets", pkts);
      for (i=0; i<pkts; i++) {
	len = *(unsigned short *)p;
	p  +=  sizeof len;
	if ( !(buffer = xkBufferPoolNextBlock(ps->buffers) ) ) {
	  xError("ethdev ERROR: Can't get buffer, dropping incoming packet");
	  xIfTraceP(self, TR_SOFT_ERRORS) {
	    xkBufferPoolDump(ps->buffers);
	  }
	}
	else {
	  memcpy(buffer->data, p, len);
	  msgTruncate(&buffer->msg, len);
	  buffer->self = self;
	  xTraceP1(self, TR_FULL_TRACE, "sem count (pre signal) == %d", buffer->sem.count);
#ifdef XK_RX_DEBUG
	  xIfTraceP(self, TR_EVENTS) {
	    ethdev_dump(self, stdout, "RX", p, len);
	  }
#endif  
	  semSignal(&buffer->sem);
	}
	p += len;
      }
    }
    Yield();			/* pass control to new packets 		*/
  }
  
  xk_master_unlock();		/* UnLock xKernel 			*/
      
  return NULL;
} /* ethdev_rxthread */



static void
ethdev_gethost(XObj self) 
{
  PState		*ps = self->state;
  int			fd;
  struct ifreq		ifr;

  if ( (fd = socket(AF_INET, SOCK_DGRAM, 0)) == -1 ) {
    xError("cannot open socket");
    perror("");
    exit(1);
  }

  strcpy(ifr.ifr_name, "eth0");
  if ( ioctl(fd, SIOCGIFHWADDR, &ifr) == -1 ) {
    xTraceP0(self, TR_ERRORS, "SIOCGIFHWADDR ioctl failed");
    exit(1);
  }

  memcpy(&ps->ethaddr, &ifr.ifr_hwaddr.sa_data, sizeof(ETHhost));
  xTraceP1(self, TR_GROSS_EVENTS, "interface address %s",
	   ethHostStr(&ps->ethaddr));

} /* ethdev_gethost */


#ifdef XK_DEBUG
static void
ethdev_dump(XObj self, FILE *fout, const char *label,
	    unsigned char *pkt, int len)
{
  PState	*ps = (PState *)self->state;
  int		i, j;

  xAssert(fout);
  xAssert(pkt);
  xAssert(label);
  xAssert(len >= 0);

	  /* Dumping is in the debug hex/ascii style and is setup for
	     80 columns.  This puts 16 bytes per line. */
 
/*   len = min(len, ps->dumplen); */
	    
  for (i=0; i<len; i+=16) {
    char	buffer[80];
    char	*p = &buffer[0];
    char	*q = &buffer[52];
    
    memset(buffer, (int)' ', sizeof buffer);
    
    for (j=0; j<16 && (j+i) < len; j++) {
      unsigned char	byte = pkt[i+j];

      if ( j == 8 )
	p += sprintf(p, "- ");
      
      p += sprintf(p, "%02x ", byte);
      
      if ( isprint(byte) )
	*q++ = byte;
      else
	*q++ = '.';
    }

    *p = ' ';
    *q = '\0';
    xTraceP2(self, TR_GROSS_EVENTS, "%s: %s", label, buffer);
  }
} /* ethdev_dump */
#endif


static xkern_return_t
ethdev_nonpromiscuous(XObj self) 
{
  PState	*ps = (PState *)self->state;
  int		sock;
  struct ifreq	ifr;

				/* This does not work with the ETHTAP
                                   module */

  /* Force the ethernet device into non-promiscuous mode.  Becuase the
   * Linux kernel does not handle the case when IFF_PROMISC changes we
   * manually take the interface down, and up with IFF_PROMISC
   * diabled.
   */
  
  if ( (sock = socket(AF_INET, SOCK_PACKET, ntohs(ETH_P_ALL))) == -1 ) {
    perror("cannot open socket");
    exit(1);
  }
  strcpy(ifr.ifr_name, "eth0");

  if ( ioctl(sock, SIOCGIFFLAGS, &ifr) == -1 ) {
    xTraceP0(self, TR_ERRORS, "ioctl SIOCGIFFLAGS failed");
    goto fail;
  }

  ifr.ifr_flags &= ~IFF_UP;
  if ( ioctl(sock, SIOCSIFFLAGS, &ifr) == -1 ) {
    xTraceP0(self, TR_ERRORS, "ioctl SIOCSIFFLAGS failed (down)");
    goto fail;
  }

  ifr.ifr_flags |= IFF_UP;
  if ( ioctl(sock, SIOCSIFFLAGS, &ifr) == -1 ) {
    xTraceP0(self, TR_ERRORS, "ioctl SIOCSIFFLAGS failed (up)");
    goto fail;
  }
  
fail:
  close(sock);
  return XK_SUCCESS;
} /* ethdev_nonpromiscuous */




/* ethdev.c ends here */
 


