/*
 * ocsum-fns.c
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Revision: 1.3 $
 * $Date: 1994/04/17 04:36:55 $
 */
/*
 * See ocsum-test.c for description.
 */
#ifdef __alpha__
#  define xk_int64	/* define if sizeof(long) == 8 */
#endif

/*
 * ocsum -- return the 1's complement sum of the 'count' 16-bit words
 * pointed to by 'hdr'.
 */
u_short
ocsum( hdr, count )
    u_short *hdr;
    int count;
{
    register u_long acc = 0;
    
    while ( count-- ) {
	acc += *hdr++;
	if ( acc & UNSIGNED(0xFFFF0000) ) {
	    /*
	     * Carry occurred -- wrap around
	     */
	    acc &= 0xFFFF;
	    acc++;
	}
    }
    return acc & UNSIGNED(0xFFFF);
} /* ocsum */


u_short
ocsum32( hdr, count )
    u_short *hdr;
    register int count;
{
    register u_int sum = 0;
    register u_int *lp;
    register u_int overflow = 0;
    
    if ((u_long)hdr & 0x2) {   /* get up to a mult of 4 address */
      sum += *hdr++;
      --count;
    }

    lp = (u_int*) hdr;
    while ( count >= 16 ) {
	sum += *lp;
	if (sum < *lp++) overflow++;
	sum += *lp;
	if (sum < *lp++) overflow++;
	sum += *lp;
	if (sum < *lp++) overflow++;
	sum += *lp;
	if (sum < *lp++) overflow++;
	sum += *lp;
	if (sum < *lp++) overflow++;
	sum += *lp;
	if (sum < *lp++) overflow++;
	sum += *lp;
	if (sum < *lp++) overflow++;
	sum += *lp;
	if (sum < *lp++) overflow++;
	count -= 16;
    } /* while */
    while ( count >= 2 ) {
	sum += *lp;
	if (sum < *lp++) overflow++;
	count -= 2;
    }
    if (count) {
      hdr = (u_short*) lp;
      sum += *hdr;
      if (sum < *hdr++) overflow++;
    }
    sum += overflow;
    if (sum < overflow) sum++;
    sum = (sum & 0xffff) + ((sum>>16) & 0xffff);
    sum = (sum & 0xffff) + ((sum>>16) & 0xffff);
    return sum;
} /* ocsum32 */


u_short
ocsum16_unrl( hdr, count )
    u_short *hdr;
    int count;
{
    u_int sum = 0;
    u_int val;
    u_int *lp;
    
    while ((u_long)hdr & 0x3) {
	sum = *hdr++;
	--count;
    }

    lp = (u_int *)hdr;
    while ( count >= 2 ) {
	val = *lp++;
	sum += val >> 16;
	if (sum > UNSIGNED(0xffff)) {
	    ++sum;
	    sum &= UNSIGNED(0xffff);
	} /* if */
	sum += val & UNSIGNED(0xffff);
	if (sum > UNSIGNED(0xffff)) {
	    ++sum;
	    sum &= UNSIGNED(0xffff);
	} /* if */
	count -= 2;
    } /* while */
    if (count) {
	hdr = (u_short*) lp;
	sum += *hdr;
	if (sum > UNSIGNED(0xffff)) {
	    ++sum;
	    sum &= UNSIGNED(0xffff);
	} /* if */
    } /* if */
    return sum;
} /* ocsum16_unrl */


u_short
ocsum16_32nocmp( hdr, count )
    u_short *hdr;
    int count;
{
    u_int sum = 0;
    u_int val;
    u_int *lp;
    
    /* align to 4 byte boundary: */
    while ((u_long)hdr & 0x3) {
	sum = *hdr++;
	--count;
    }

    lp = (u_int *)hdr;
    while ( count >= 2 ) {
	val = *lp++;
	sum += val >> 16;
	sum += val & UNSIGNED(0xffff);
	count -= 2;
    } /* while */
    if (count) {
	sum += *(u_short *)lp;
    } /* if */
    if (sum > 0xffff) {
	sum = (sum & UNSIGNED(0xffff)) + ((sum>>16) & UNSIGNED(0xffff));
	sum = (sum & UNSIGNED(0xffff)) + ((sum>>16) & UNSIGNED(0xffff));
    }
    return sum;
} /* ocsum16_32nocmp */


u_short
ocsum16_32nocmp_unrl( hdr, count )
    u_short *hdr;
    int count;
{
    u_int sum = 0;
    u_int val;
    u_int *lp;
    
    while ((u_long)hdr & 0x3) {
	sum = *hdr++;
	--count;
    }

    lp = (u_int *)hdr;
    while ( count >= 8 ) {
	val = *lp++;
	sum += val >> 16;
	sum += val & UNSIGNED(0xffff);
	val = *lp++;
	sum += val >> 16;
	sum += val & UNSIGNED(0xffff);
	val = *lp++;
	sum += val >> 16;
	sum += val & UNSIGNED(0xffff);
	val = *lp++;
	sum += val >> 16;
	sum += val & UNSIGNED(0xffff);
	count -= 8;
      } /* while */
    while (count>=2) {
	val = *lp++;
	sum += val >> 16;
	sum += val & UNSIGNED(0xffff);
	count -= 2;
      } /* if */
    if (count) sum += *(u_short *)lp;

    if (sum > 0xffff) {
      sum = (sum & UNSIGNED(0xffff)) + ((sum>>16) & UNSIGNED(0xffff));
      sum = (sum & UNSIGNED(0xffff)) + ((sum>>16) & UNSIGNED(0xffff));
    }
    return sum;
} /* ocsum16_32nocmp_unrl */


u_short
ocsum_nocmp( hdr, count )
    u_short *hdr;
    int count;
{
/* 
 * ccsum.c - Highly optimized MIPS checksum function.
 * by Jonathan Kay, Computer Systems Lab, UCSD         4/2/93
 *
 * Version 2.0
 * Techniques and credits:
 *   Basic algorithm is 3-instruction inner loop sum by Peter Desnoyers.
 *   Full word-size reading as described in Usenix W'93 paper.
 *   Pipelined latency absoption technique as described in paper.
 *   Unrolling chosen through testing and examination of actual workload.
 *   Rewrite in 'C' without loss of performance suggested by Vernon Schryver.
 *   15% faster than version 1 ("Usenix version").
 *   150% faster than Ultrix 4.2A checksum routine.
 */

/*
 * Copyright (c) 1993 Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * paragraph appears in all copies of this software.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
 * EVENT SHALL THE REGENTS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

    u_long hilo, high;
    u_long w0, w1;
    union memptr {
	u_int *i;
	u_long *l;
	u_long u;
	u_short *s;
	u_char *c;
    } buf;
    u_long sum = 0;
#   define HALF_REG_WIDTH	(8 * sizeof(u_long) / 2)
#   define SHORTS_PER_LONG	(sizeof(long)/sizeof(short))

    buf.s = hdr;
	
#ifdef xk_int64
    /* 64-bit align: */
    switch (buf.u & 0x6) {
      case 2: sum += *buf.s++; if (--count == 0) break;
      case 4: sum += *buf.s++; if (--count == 0) break;
      case 6: sum += *buf.s++; if (--count == 0) break;
      case 0: break;
    } /* switch */
#else
    /* 32-bit-align */
    if (buf.u & 0x2) {
	sum += *(buf.s++);
	--count;
    }
#endif /* xk_int64 */
	
    /*
     * long-aligned sum, 16 times unrolled.
     * Peter Desnoyers' unbelievable 3-instruction main loop.
     */
    if (count >= (16 + 2) * SHORTS_PER_LONG) {
	w0 = buf.l[0];
	w1 = buf.l[1];
	high = hilo = 0;
	do {
	    count -= 16 * SHORTS_PER_LONG;
	    hilo += w0; high += w0 >> HALF_REG_WIDTH; w0 = buf.l[2];
	    hilo += w1; high += w1 >> HALF_REG_WIDTH; w1 = buf.l[3];
	    hilo += w0; high += w0 >> HALF_REG_WIDTH; w0 = buf.l[4];
	    hilo += w1; high += w1 >> HALF_REG_WIDTH; w1 = buf.l[5];
	    hilo += w0; high += w0 >> HALF_REG_WIDTH; w0 = buf.l[6];
	    hilo += w1; high += w1 >> HALF_REG_WIDTH; w1 = buf.l[7];
	    hilo += w0; high += w0 >> HALF_REG_WIDTH; w0 = buf.l[8];
	    hilo += w1; high += w1 >> HALF_REG_WIDTH; w1 = buf.l[9];

	    hilo += w0; high += w0 >> HALF_REG_WIDTH; w0 = buf.l[10];
	    hilo += w1; high += w1 >> HALF_REG_WIDTH; w1 = buf.l[11];
	    hilo += w0; high += w0 >> HALF_REG_WIDTH; w0 = buf.l[12];
	    hilo += w1; high += w1 >> HALF_REG_WIDTH; w1 = buf.l[13];
	    hilo += w0; high += w0 >> HALF_REG_WIDTH; w0 = buf.l[14];
	    hilo += w1; high += w1 >> HALF_REG_WIDTH; w1 = buf.l[15];
	    hilo += w0; high += w0 >> HALF_REG_WIDTH; w0 = buf.l[16];
	    hilo += w1; high += w1 >> HALF_REG_WIDTH; w1 = buf.l[17];
	    buf.l += 16;
	} while (count >= (16 + 2) * SHORTS_PER_LONG);
	hilo -= (high << HALF_REG_WIDTH);
	sum += hilo;
	sum += high;
    } /* if */
    high = hilo = 0;
    while (count >= SHORTS_PER_LONG) {
	count -= SHORTS_PER_LONG;
	w0 = *(buf.l++); hilo += w0; high += w0 >> HALF_REG_WIDTH;
    } /* while */
    hilo -= (high << HALF_REG_WIDTH);
    sum += hilo;
    sum += high;

    while (count) {
	--count;
	sum += *(buf.s++);
    } /* while */
	
    /* add all shorts in a long together to get full sum: */
#ifdef xk_int64
    sum = (sum & 0xffff) + ((sum>>16) & 0xffff) +
      ((sum>>32) & 0xffff) + ((sum>>48) & 0xffff);
#else
    sum = (sum & 0xffff) + (sum >> 16);
#endif /* xk_int64 */
    sum = (sum & 0xffff) + (sum >> 16);
    return sum;
#   undef HALF_REG_WIDTH
#   undef SHORTS_PER_LONG
} /* ocsum_nocmp */


/*
 * Same as ocsum_nocmp except that there is an additional loop that
 * is unrolled twice (to process the remainder/small packets quickly).
 */
u_short
ocsum_nocmp_small( hdr, count )
    u_short *hdr;
    int count;
{
/* 
 * ccsum.c - Highly optimized MIPS checksum function.
 * by Jonathan Kay, Computer Systems Lab, UCSD         4/2/93
 *
 * Version 2.0
 * Techniques and credits:
 *   Basic algorithm is 3-instruction inner loop sum by Peter Desnoyers.
 *   Full word-size reading as described in Usenix W'93 paper.
 *   Pipelined latency absoption technique as described in paper.
 *   Unrolling chosen through testing and examination of actual workload.
 *   Rewrite in 'C' without loss of performance suggested by Vernon Schryver.
 *   15% faster than version 1 ("Usenix version").
 *   150% faster than Ultrix 4.2A checksum routine.
 */

/*
 * Copyright (c) 1993 Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * paragraph appears in all copies of this software.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
 * EVENT SHALL THE REGENTS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

    u_long hilo, high;
    u_long w0, w1;
    union memptr {
	u_int *i;
	u_long *l;
	u_long u;
	u_short *s;
	u_char *c;
    } buf;
    u_long sum = 0;
#   define HALF_REG_WIDTH	(8 * sizeof(u_long) / 2)
#   define SHORTS_PER_LONG	(sizeof(long)/sizeof(short))

    buf.s = hdr;
	
#ifdef xk_int64
    /* 64-bit align: */
    switch (buf.u & 0x6) {
      case 2: sum += *buf.s++; if (--count == 0) break;
      case 4: sum += *buf.s++; if (--count == 0) break;
      case 6: sum += *buf.s++; if (--count == 0) break;
      case 0: break;
    } /* switch */
#else
    /* 32-bit-align */
    if (buf.u & 0x2) {
	sum += *(buf.s++);
	--count;
    }
#endif /* xk_int64 */
	
    /*
     * long-aligned sum, 16 times unrolled.
     * Peter Desnoyers' unbelievable 3-instruction main loop.
     */
    if (count >= (16 + 2) * SHORTS_PER_LONG) {
	w0 = buf.l[0];
	w1 = buf.l[1];
	high = hilo = 0;
	do {
	    hilo += w0;
	    high += w0 >> HALF_REG_WIDTH;
	    w0 = buf.l[2];

	    hilo += w1;
	    high += w1 >> HALF_REG_WIDTH;
	    w1 = buf.l[3];

	    hilo += w0;
	    high += w0 >> HALF_REG_WIDTH;
	    w0 = buf.l[4];

	    hilo += w1;
	    high += w1 >> HALF_REG_WIDTH;
	    w1 = buf.l[5];

	    hilo += w0;
	    high += w0 >> HALF_REG_WIDTH;
	    w0 = buf.l[6];

	    hilo += w1;
	    high += w1 >> HALF_REG_WIDTH;
	    w1 = buf.l[7];

	    hilo += w0;
	    high += w0 >> HALF_REG_WIDTH;
	    w0 = buf.l[8];

	    hilo += w1;
	    high += w1 >> HALF_REG_WIDTH;
	    w1 = buf.l[9];


	    hilo += w0;
	    high += w0 >> HALF_REG_WIDTH;
	    w0 = buf.l[10];

	    hilo += w1;
	    high += w1 >> HALF_REG_WIDTH;
	    w1 = buf.l[11];

	    hilo += w0;
	    high += w0 >> HALF_REG_WIDTH;
	    w0 = buf.l[12];

	    hilo += w1;
	    high += w1 >> HALF_REG_WIDTH;
	    w1 = buf.l[13];

	    hilo += w0;
	    high += w0 >> HALF_REG_WIDTH;
	    w0 = buf.l[14];

	    hilo += w1;
	    high += w1 >> HALF_REG_WIDTH;
	    w1 = buf.l[15];

	    hilo += w0;
	    high += w0 >> HALF_REG_WIDTH;
	    w0 = buf.l[16];

	    hilo += w1;
	    high += w1 >> HALF_REG_WIDTH;
	    w1 = buf.l[17];

	    count -= 16 * SHORTS_PER_LONG;
	    buf.l += 16;
	} while (count >= (16 + 2) * SHORTS_PER_LONG);
	hilo -= (high << HALF_REG_WIDTH);
	sum += hilo;
	sum += high;
    } /* if */
    /*
     * long-aligned sum, 2 times unrolled.
     */
    if (count >= (2 + 2) * SHORTS_PER_LONG) {
	w0 = buf.l[0];
	w1 = buf.l[1];
	high = hilo = 0;
	do {
	    hilo += w0;
	    high += w0 >> HALF_REG_WIDTH;
	    w0 = buf.l[2];

	    hilo += w1;
	    high += w1 >> HALF_REG_WIDTH;
	    w1 = buf.l[3];

	    count -= 2 * SHORTS_PER_LONG;
	    buf.l += 2;
	} while (count >= (2 + 2) * SHORTS_PER_LONG);
	hilo -= (high << HALF_REG_WIDTH);
	sum += hilo;
	sum += high;
    } /* if */
    high = hilo = 0;
    while (count >= SHORTS_PER_LONG) {
	count -= SHORTS_PER_LONG;
	w0 = *(buf.l++);
	hilo += w0;
	high += w0 >> HALF_REG_WIDTH;
    } /* while */
    hilo -= (high << HALF_REG_WIDTH);
    sum += hilo;
    sum += high;

    while (count) {
	--count;
	sum += *(buf.s++);
    } /* while */
	
    /* add all shorts in a long together to get full sum: */
#ifdef xk_int64
    sum = (sum & 0xffff) + ((sum>>16) & 0xffff) +
      ((sum>>32) & 0xffff) + ((sum>>48) & 0xffff);
#else
    sum = (sum & 0xffff) + (sum >> 16);
#endif /* xk_int64 */
    sum = (sum & 0xffff) + (sum >> 16);
    return sum;
#   undef HALF_REG_WIDTH
#   undef SHORTS_PER_LONG
} /* ocsum_nocmp_small */


/*
 * Returns the one's complement sum of the COUNT 16-bit words pointed
 * to by HDR.  HDR must point to an even address.
 */
u_short
ocsum_asm(u_short *sp, int count)
{
    u_long sum = 0;
    u_long *lp;
    u_long overflow = 0;
    u_long t1, t2, t3, t4, t5, t6, t7, t8;
#   define SHORTS_PER_LONG	(sizeof(long)/sizeof(short))
    
    /* make SP int aligned: */

#if defined(xk_int64) || defined(__sparc__)

    switch ((u_long) sp & 0x6) {
      case 2: sum += *sp++; if (--count == 0) break;
      case 4: sum += *sp++; if (--count == 0) break;
      case 6: sum += *sp++; if (--count == 0) break;
      case 0: break;
    } /* switch */

#else

    if ((u_long)sp & 2) {
	sum += *sp++; --count;
    } /* if */

#endif /* xk_int64 */

    /* move long word by long word, eight times unrolled: */
    lp = (u_long*) sp;
#if defined(__GNUC__) && (defined(__hppa__) || defined(__sparc__))
# if __hppa__
    while (count >= 16 * SHORTS_PER_LONG) {
	u_long t0, t1, t2, t3, t4, t5, t6, t7;

	asm volatile ("ldw 0(%1),%0" : "=r"(t0)  : "r"(lp));
	asm volatile ("ldw 4(%1),%0" : "=r"(t1)  : "r"(lp));
	asm volatile ("ldw 8(%1),%0" : "=r"(t2)  : "r"(lp));
	asm volatile ("add  %2,%1,%0" : "=r"(sum) : "r"(t0), "r"(sum));
	asm volatile ("ldw 12(%1),%0" : "=r"(t3)  : "r"(lp));
	asm volatile ("addc %2,%1,%0" : "=r"(sum) : "r"(t1), "r"(sum));
	asm volatile ("ldw 16(%1),%0" : "=r"(t4)  : "r"(lp));
	asm volatile ("addc %2,%1,%0" : "=r"(sum) : "r"(t2), "r"(sum));
	asm volatile ("ldw 20(%1),%0" : "=r"(t5)  : "r"(lp));
	asm volatile ("addc %2,%1,%0" : "=r"(sum) : "r"(t3), "r"(sum));
	asm volatile ("ldw 24(%1),%0" : "=r"(t6)  : "r"(lp));
	asm volatile ("addc %2,%1,%0" : "=r"(sum) : "r"(t4), "r"(sum));
	asm volatile ("ldw 28(%1),%0" : "=r"(t7)  : "r"(lp));
	asm volatile ("addc %2,%1,%0" : "=r"(sum) : "r"(t5), "r"(sum));
	asm volatile ("ldw 32(%1),%0" : "=r"(t0)  : "r"(lp));
	asm volatile ("addc %2,%1,%0" : "=r"(sum) : "r"(t6), "r"(sum));
	asm volatile ("ldw 36(%1),%0" : "=r"(t1)  : "r"(lp));
	asm volatile ("addc %2,%1,%0" : "=r"(sum) : "r"(t7), "r"(sum));
	asm volatile ("ldw 40(%1),%0" : "=r"(t2)  : "r"(lp));

	asm volatile ("addc %2,%1,%0" : "=r"(sum) : "r"(t0), "r"(sum));
	asm volatile ("ldw 44(%1),%0" : "=r"(t3)  : "r"(lp));
	asm volatile ("addc %2,%1,%0" : "=r"(sum) : "r"(t1), "r"(sum));
	asm volatile ("ldw 48(%1),%0" : "=r"(t4)  : "r"(lp));
	asm volatile ("addc %2,%1,%0" : "=r"(sum) : "r"(t2), "r"(sum));
	asm volatile ("ldw 52(%1),%0" : "=r"(t5)  : "r"(lp));
	asm volatile ("addc %2,%1,%0" : "=r"(sum) : "r"(t3), "r"(sum));
	asm volatile ("ldw 56(%1),%0" : "=r"(t6)  : "r"(lp));
	asm volatile ("addc %2,%1,%0" : "=r"(sum) : "r"(t4), "r"(sum));
	asm volatile ("ldw 60(%1),%0" : "=r"(t7)  : "r"(lp));
	asm volatile ("addc %2,%1,%0" : "=r"(sum) : "r"(t5), "r"(sum));
	asm volatile ("addc %2,%1,%0" : "=r"(sum) : "r"(t6), "r"(sum));
	asm volatile ("addc %2,%1,%0" : "=r"(sum) : "r"(t7), "r"(sum));
	asm volatile ("addc %1,0,%0"  : "=r"(sum) : "r"(sum));
	lp += 16;
	count -= 16 * SHORTS_PER_LONG;
    } /* while */
# endif /* __hppa__ */
# if __sparc__
    while (count >= 16 * SHORTS_PER_LONG) {
	asm volatile ("ldd [%0],%%l0"     :	      :"r"(lp) : "%l0", "%l1");
	asm volatile ("ldd [%0+8],%%l2"   :	      :"r"(lp) : "%l2", "%l3");
	asm volatile ("addcc  %1,%%l0,%0" : "=r"(sum) :"r"(sum));
	asm volatile ("addxcc %1,%%l1,%0" : "=r"(sum) :"r"(sum));
	asm volatile ("ldd [%0+16],%%l0"  :	      :"r"(lp) : "%l0", "%l1");
	asm volatile ("addxcc %1,%%l2,%0" : "=r"(sum) :"r"(sum));
	asm volatile ("addxcc %1,%%l3,%0" : "=r"(sum) :"r"(sum));
	asm volatile ("ldd [%0+24],%%l2"  :	      :"r"(lp) : "%l2", "%l3");
	asm volatile ("addxcc %1,%%l0,%0" : "=r"(sum) :"r"(sum));
	asm volatile ("addxcc %1,%%l1,%0" : "=r"(sum) :"r"(sum));
	asm volatile ("ldd [%0+32],%%l0"  :	      :"r"(lp) : "%l0", "%l1");
	asm volatile ("addxcc %1,%%l2,%0" : "=r"(sum) :"r"(sum));
	asm volatile ("addxcc %1,%%l3,%0" : "=r"(sum) :"r"(sum));

	asm volatile ("ldd [%0+40],%%l2"  :	      :"r"(lp) : "%l2", "%l3");
	asm volatile ("addxcc %1,%%l0,%0" : "=r"(sum) :"r"(sum));
	asm volatile ("addxcc %1,%%l1,%0" : "=r"(sum) :"r"(sum));
	asm volatile ("ldd [%0+48],%%l0"  :	      :"r"(lp) : "%l0", "%l1");
	asm volatile ("addxcc %1,%%l2,%0" : "=r"(sum) :"r"(sum));
	asm volatile ("addxcc %1,%%l3,%0" : "=r"(sum) :"r"(sum));
	asm volatile ("ldd [%0+56],%%l2"  :	      :"r"(lp) : "%l2", "%l3");
	asm volatile ("addxcc %1,%%l0,%0" : "=r"(sum) :"r"(sum));
	asm volatile ("addxcc %1,%%l1,%0" : "=r"(sum) :"r"(sum));
	asm volatile ("addxcc %1,%%l2,%0" : "=r"(sum) :"r"(sum));
	asm volatile ("addxcc %1,%%l3,%0" : "=r"(sum) :"r"(sum));
	asm volatile ("addx   %1,0,%0"    : "=r"(sum) :"r"(sum));
	lp += 16;
	count -= 16 * SHORTS_PER_LONG;
    } /* while */
# endif
#else
    if (count >= (8 + 1) * SHORTS_PER_LONG) {
	t1 = lp[0]; 
	do {
	    t5 = lp[4]; t2 = lp[1]; t3 = lp[2]; t4 = lp[3];
	    sum += t1; if (sum < t1) overflow++;
	    sum += t2; if (sum < t2) overflow++;
	    sum += t3; if (sum < t3) overflow++;
	    sum += t4; if (sum < t4) overflow++;

	    count -= 8 * SHORTS_PER_LONG;

	    t1 = lp[8]; t6 = lp[5]; t7 = lp[6]; t8 = lp[7];
	    lp += 8;
	    sum += t5; if (sum < t5) overflow++;
	    sum += t6; if (sum < t6) overflow++;
	    sum += t7; if (sum < t7) overflow++;
	    sum += t8; if (sum < t8) overflow++;
	} while (count >= (8 + 1) * SHORTS_PER_LONG);
    } /* if */
#endif
    while (count >= SHORTS_PER_LONG) {
	count -= SHORTS_PER_LONG;
	sum += *lp; if (sum < *lp++) ++overflow;
    } /* while */
    if (count) {
	sp = (u_short*) lp;
	do {
	    --count;
	    sum += *sp; if (sum < *sp++) ++overflow;
	} while (count);
    } /* if */
    sum += overflow; if (sum < overflow) sum++;

#ifdef xk_int64
    sum = (sum & 0xffff) + ((sum>>16) & 0xffff) +
      ((sum>>32) & 0xffff) + ((sum>>48) & 0xffff);
#else
    sum = (sum & 0xffff) + ((sum>>16) & 0xffff);
#endif
    sum = (sum & 0xffff) + ((sum>>16) & 0xffff);
    return sum;
#   undef SHORTS_PER_LONG
} /* ocsum_asm */


#ifdef xk_int64

u_short
ocsum32_64nocmp( hdr, count )
     u_short *hdr;
     int count;
{
    u_long sum = 0;
    u_long val;
    u_long *lp;
    u_short *sp;

    /* align to 8 byte boundary: */
    switch ((u_long) hdr & 0x6) {
      case 2: sum += *hdr++; if (--count == 0) break;
      case 4: sum += *hdr++; if (--count == 0) break;
      case 6: sum += *hdr++; if (--count == 0) break;
      case 0: break;
    } /* switch */

    lp = (u_long*)hdr;
    while (count >= 4) {
	val = *lp++;
	sum += val >> 32;
	sum += val & 0xffffffffL;
	count -= 4;
    } /* while */
    if (count) {
	sp = (u_short*) lp;
	while (count--) {
	    sum += *sp++;
	} /* while */
    } /* if */
    sum = (sum & 0xffff) + ((sum>>16) & 0xffff) +
      ((sum>>32) & 0xffff) + ((sum>>48) & 0xffff);
    sum = (sum & 0xffff) + (sum >> 16);
    return sum;
} /* ocsum32_64nocmp */


u_short
ocsum32_64nocmp_unrl( hdr, count )
     u_short *hdr;
     int count;
{
    u_long sum = 0;
    u_long w0, w1;
    u_long *lp;
    u_short *sp;
#   define SHORTS_PER_LONG	(sizeof(long)/sizeof(short))

    /* align to 8 byte boundary: */
    switch ((u_long) hdr & 0x7) {
      case 2: sum += *hdr++; if (--count == 0) break;
      case 4: sum += *hdr++; if (--count == 0) break;
      case 6: sum += *hdr++; if (--count == 0) break;
      case 0: break;
    } /* switch */

    lp = (u_long*)hdr;

    if (count >= (4 + 2) * SHORTS_PER_LONG) {
	w0 = lp[0]; w1 = lp[1];
	do {
	    count -= 4 * SHORTS_PER_LONG;
	    sum += w0 >> 32; sum += w0 & 0xffffffffL; w0 = lp[2];
	    sum += w1 >> 32; sum += w1 & 0xffffffffL; w1 = lp[3];
	    sum += w0 >> 32; sum += w0 & 0xffffffffL; w0 = lp[4];
	    sum += w1 >> 32; sum += w1 & 0xffffffffL; w1 = lp[5];
	    lp += 4;
	} while (count >= (4 + 2) * SHORTS_PER_LONG);
    } /* if */
    while (count >= SHORTS_PER_LONG) {
	count -= SHORTS_PER_LONG;
	w0 = *lp++; sum += w0 >> 32; sum += w0 & 0xffffffffL;
    } /* while */
    if (count) {
	sp = (u_short*) lp;
	while (count--) {
	    sum += *sp++;
	} /* while */
    } /* if */
    sum = (sum & 0xffff) + ((sum>>16) & 0xffff) +
      ((sum>>32) & 0xffff) + ((sum>>48) & 0xffff);
    sum = (sum & 0xffff) + (sum >> 16);
    return sum;
#   undef SHORTS_PER_LONG
} /* ocsum32_64nocmp_unrl */

#endif /* xk_int64 */

static struct {
    u_short (*fun)();
    char    *name;
} ocsum_fns[] = {
    /* first entry should be slow but correct function: */
    {ocsum,			"ocsum"},
    {ocsum16_unrl,		"ocsum16_unrl"},
    {ocsum16_32nocmp,		"ocsum16_32nocmp"},
    {ocsum16_32nocmp_unrl,	"ocsum16_32nocmp_unrl"},
    {ocsum32,			"ocsum32"},
#ifdef xk_int64
    {ocsum32_64nocmp,		"ocsum32_64nocmp"},
    {ocsum32_64nocmp_unrl,	"ocsum32_64nocmp_unrl"},
#endif
    {ocsum_nocmp,		"ocsum_nocmp"},
    {ocsum_nocmp_small,		"ocsum_nocmp_small"},
    {ocsum_asm,			"ocsum_asm"},
};
