#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <pwd.h>
#include <ctype.h>
#include "xkernel.h"
#include <vtap.h>
#include <saidmgr.h>
#include "tcp.h"
#include "eth.h"
#include "ip.h"
#undef IPPORT_LOGINSERVER
#include "rl_i.h"

/* log:
 * 	used popen()
 *	used pipe()
 *	used socketpair() to get better buffering
 */


#define DEFAULT_SERV_ADDR "128.10.5.55"
#define DEFAULT_TERM_TYPE "xterm"
#define DEFAULT_USER "guest"

#define RHOSTS_FILE ".rhosts"

#define SHELL_BUFFERSZ 8192

void proxyrl_init(XObj self);
void client_init(Event e, void *tself);
void client_kbdpoll(Event e, void *tself);
void client_scripthandle(Event e, void *tself);
static xkern_return_t proxyrlDemux(XObj self, XObj lls, Msg * msg);
void proxyrl_msgtest(void);
int s_pipe(int fd[2]);

#if defined(__STDC__) || defined(__cplusplus)
#define P_(s) s
#else
#define P_(s) ()
#endif

#ifdef SUNOS
#include <unistd.h>
#include <memory.h>
#include <ctype.h>
void perror P_((char *));
int select P_((int, fd_set*, fd_set*, fd_set*, const struct timeval*));
int socketpair P_((int, int, int, int *));
int setsockopt P_((int, int, int, char *, int));
int system P_((char *));
int fputs P_((char *, FILE *));
#endif


#ifdef SOLARIS
#include <unistd.h>
#endif

/* proxyrl/rl.c */
static char * mystrdup(char *);
static long mycopypop P_((void *dest, char *src, long count, void *arg));
/*static void mycopypush P_((void *src, char *dst, long count, void *arg));*/
static xkern_return_t proxyrlPop P_((XObj s, XObj lls, Msg *msg, VOID *arg));
void proxyrl_init P_((XObj self));
static void sessnInit P_((XObj s));
static void protInit P_((XObj p));
static XObj proxyrlOpen P_((XObj self, XObj hlpRcv, XObj hlpType, Part *p));
/*static xkern_return_t proxyrlOpenEnable P_((XObj self, XObj hlpRcv, XObj hlpType, Part *p));*/
static XObj proxyrlCreateSession P_((XObj self, XObj hlpRcv, XObj hlpType, XObj lls));
void server_rhosts_load P_((void));
int hostname_islegal P_((char *hostname));
void client_init P_((Event e, void *tself));
void client_kbdpoll P_((Event e, void *tself));
static xkern_return_t proxyrlDemux P_((XObj self, XObj lls, Msg *msg));
int string_split_space P_((char *string, char *s, char *t));
static int proxyrlControlSessn( XObj self, int opcode, char *buf, int len);
static int proxyrlControlProtl( XObj self, int opcode, char *buf, int len);

int traceproxyrlp = 0;
int traceproxyrl = 0;
int isserver = 0;
Msg proxyrl_msg;
int spoofmode = 0;
char *pcspoofaddr = NULL;
int vtapmode = 0;

typedef struct {
    char *infile;
    char *outfile;
} script;

script *pscripts = NULL;		/* array of scripts */
int numscripts = 0;

static char * mystrdup(char * inp)
{
	char * buf;
	buf = xMalloc(strlen(inp)+1);
	strcpy(buf,inp);
	return buf;
}

static long mycopypop(dest, src, count, arg)
     void *dest;
     char *src;
     long count;
     void *arg;
{
    bcopy(src, (char*)dest, count);
    return count;
}

/*
static void mycopypush(src, dst, count, arg)
     void *src;
     char *dst;
     long  count;
     void *arg;
{
    bcopy((char *)src, dst, count);
}
*/


/* proxyrl_init: creates map, sets up functions
 */
void 
proxyrl_init(XObj self)
{
    PState *ps;
    char *pc;
    int i;
    XObj proxyrl_llp;
    int la, lb, lc, ld;

    xTrace0(proxyrlp, TR_GROSS_EVENTS, "PROXYRL init");
    xAssert(xIsProtocol(self));

    if (!xIsProtocol(proxyrl_llp = xGetDown(self, 0))) {
	Kabort("PROXYRL could not get lower protocol");
    }

    /* make sure we can get all messages addressed to the spoofed host */
    if (spoofmode) {
	xControl(proxyrl_llp, ETH_SETPROMISCUOUS, 0, 0);
    }

    /* initialize all those data structures associated with XObjs */
    protInit(self);

    /* set up the maps */
    ps = X_NEW(PState);
    self->state = (VOID *)ps;
    ps->activeMap = mapCreate(RLOGIN_ACT_MAP_SZ, sizeof(XObj));
    ps->passiveMap = mapCreate(RLOGIN_PAS_MAP_SZ, sizeof(XObj));
    ps->algorithm = NULL;
    ps->spi = 0;
    ps->mode = 0;
    str2ipHost(&ps->s_iphost,DEFAULT_SERV_ADDR);
    ps->termtype = mystrdup(DEFAULT_TERM_TYPE);
    if (self->instName[0] != 0)
	ps->locuser = mystrdup(self->instName);
    else
    {
	struct passwd * user;
	user = getpwuid(getuid());
	if (user != NULL)
		ps->locuser = mystrdup(user->pw_name);
	else
		ps->locuser = mystrdup(DEFAULT_USER); /*??? (os)*/
    }
	
    /* if not set otherwise in command line remuser == locuser (os)*/
    ps->remuser = mystrdup(ps->locuser);

    /* parse arguments */
    for (i = 1; i < globalArgc; i++) {
	pc = globalArgv[i];
        if (strcmp(pc, "-s") == 0) {
	    isserver = 1;
	    continue;
	}
        if (strncmp(pc, "-c", 2) == 0) {
	    isserver = 0;
	    if (pc[2]) {
		str2ipHost(&ps->s_iphost,pc+2);
		printf("---> server addr: %s\n", pc+2);
	    }
	    continue;
	}
        if (strcmp(pc, "-v") == 0) {
	    vtapmode = 1;
	    continue;
	}
        if (strncmp(pc, "-a", 2) == 0) {
	    if (pc[2]) {
		ps->algorithm = mystrdup(pc+2);
		printf("---> algorithm : %s\n", ps->algorithm);
	    }
	    continue;
	}
         if (strncmp(pc, "-m", 2) == 0) {
	    if (pc[2]) {
		ps->mode = atoi((char *) &pc[2]);
		printf("---> mode : %u\n", ps->mode);
	    }
	    continue;
	}
        if (strncmp(pc, "-spi", 4) == 0) {
	    if (pc[4]) {
		int x = 4;
		while((pc[x] != 0) && (pc[x] != ',')) x++;
		if (pc[x] != 0)
		{
			pc[x] = 0;
			if (sscanf(pc+(x+1),"%u.%u.%u.%u",&la,&lb,&lc,&ld) == 4)
			{
				ps->spi = atoi(pc+4);
				ps->myhost.a=la; ps->myhost.b=lb;
				ps->myhost.c=lc; ps->myhost.d=ld;
				printf("---> using local SPI and address : %u %u.%u.%u.%u\n", 
					ps->spi,la,lb,lc,ld);
			}
		}
	    }
	    continue;
	}
        if (strncmp(pc, "-ru", 3) == 0) {
	    if (pc[3]) {
		xFree(ps->remuser);
		ps->remuser = mystrdup(pc+3);
		printf("---> remote user : %s\n", ps->remuser);
	    }
	    continue;
	}
        if (strncmp(pc, "-lu", 3) == 0) {
	    if (pc[3]) {
		xFree(ps->locuser);
		ps->locuser = mystrdup(pc+3);
		printf("---> local user : %s\n", ps->locuser);
	    }
	    continue;
	}
        if (strncmp(pc, "-t", 2) == 0) {
	    if (pc[2]) {
		xFree(ps->termtype);
		ps->termtype = mystrdup(pc+2);
		printf("---> termtype : %s\n", ps->termtype);
	    }
	    continue;
	}
	if (strcmp(pc, "-spoof") == 0) {
	    spoofmode = 1;
	    pcspoofaddr = globalArgv[++i];
	    continue;
	}
	if (strcmp(pc, "-script") == 0) {
	    /* allocate new space */
	    if (numscripts == 0) {
	        if (!(pscripts = calloc(1, sizeof(script)))) {
		    perror("calloc()");
		    xAssert(0);
		} 
	    } else {
	        if (!(pscripts = realloc(pscripts,(numscripts+1)*sizeof(script)))) {
		    perror("calloc()");
		    xAssert(0);
		} 
	    }
	    pc = globalArgv[++i];
	    pscripts[numscripts].infile = pc;
	    pc = globalArgv[++i];
	    pscripts[numscripts].outfile = pc;

	    /* check arguments */
	    if (!pscripts[numscripts].outfile || !pscripts[numscripts].outfile) {
	        fprintf(stderr, "invalid argument for -script switch!\n");
		xAssert(0);
	    }
	    numscripts++;
	    continue;
	}
    }
 

    /* server code */
    if (isserver) {
        Kabort("PROXYRL does not support server mode!");
    } 
    /* client code */
    else {
	xTrace0(proxyrlp, TR_GROSS_EVENTS, "PROXYRL init: I am the client!");
	evDetach(evSchedule(client_init, self, 1000));
    }
    xTrace0(proxyrlp, TR_GROSS_EVENTS, "PROXYRL init: done");
}					/* end proxyrl_init */

static XObj
proxyrlOpen( self, hlpRcv, hlpType, p )
    XObj	self, hlpRcv, hlpType;
    Part 	*p;
{
    XObj	s, lls;
    PState	*ps = (PState *)self->state;
    
    xTrace0(proxyrlp, TR_MAJOR_EVENTS, "PROXYRL open");
    if ( (lls = xOpen(self, hlpType, xGetDown(self, 0), p)) == ERR_XOBJ ) {
	xTrace0(proxyrlp, TR_ERRORS, "proxyrlOpen: could not open lower session");
	return ERR_XOBJ;
    }
    if ( mapResolve(ps->activeMap, &lls, &s) == XK_SUCCESS ) {
	xTrace0(proxyrlp, TR_MORE_EVENTS, "proxyrlOpen -- found an existing one");
    } else {
	xTrace0(proxyrlp, TR_MAJOR_EVENTS, "proxyrlOpen -- creating a new one");
	if ((s = proxyrlCreateSession(self, hlpRcv, hlpType, lls)) == ERR_XOBJ) {
	    xClose(lls);
	}
    }
    xDuplicate(s);

    /* do we do address spoofing? */
    if (spoofmode) {
	IPhost spoofaddr;

	xTrace1(proxyrlp, TR_EVENTS, "PROXYRL spoofing address %s", pcspoofaddr);
	str2ipHost(&spoofaddr, pcspoofaddr);
	if (xControl(lls, IP_SETSOURCEADDR, (char *) &spoofaddr, sizeof(spoofaddr)) < 0) {
	    xError("PROXYRL: IP_SETSOURCEADDR failed!");
	}
    }
    return s;
}

static XObj
proxyrlCreateSession(self, hlpRcv, hlpType, lls)
    XObj        self, hlpRcv, hlpType, lls;
{
    XObj s;
    PState *ps = (PState *)self->state;
    SState *ss;

    if ((s = xCreateSessn(sessnInit, hlpRcv, hlpType, self, 1, &lls))
                        == ERR_XOBJ ) {
        xTrace0(proxyrlp, TR_ERRORS, "create sessn fails in proxyrlCreateSession");
        return ERR_XOBJ;
    }
    s->binding = mapBind(ps->activeMap, &lls, s);
    if (s->binding == ERR_BIND) {
        xTrace0(proxyrlp, TR_ERRORS, "mapBind fails in vtapCreateSessn");
        return ERR_XOBJ;
    }
    ss = X_NEW(SState);
    xAssert((ss->buf = malloc(SHELL_BUFFERSZ)) != NULL);
    s->state = (VOID *)ss;

    xTrace1(proxyrlp, TR_MAJOR_EVENTS, "PROXYRL open returns %x", s);
    return s;
}


void 
client_init(Event e, void *tself)
{
    Part plist[3];
    XObj self, prlsession;
    static long s_port = IPPORT_LOGINSERVER, c_port = ANY_PORT;
    XObj proxyrl_llp;
    int i;
    PState *ps;
    SState *ss;
    XSaid	said;

    self = (XObj) tself;
    ps = self->state;

    xTrace0(proxyrlp, TR_GROSS_EVENTS, "PROXYRL client_init: starting");

    xAssert(xIsXObj(self));
    if (!xIsProtocol(proxyrl_llp = xGetDown(self, 0))) {
	Kabort("client_init: PROXYRL could not get lower protocol");
    }


    /* foreach kbd, script {
     * 	   do the xOpen()
     *	   evDetach a thread to handle the session
     * }
     */

    for (i = -1; i < numscripts; i++) {

	/* regenerate the participant list each time */
	if (( ps->algorithm != NULL) || ( ps->spi != 0 ))
		partInit(plist, 3);
	else
		partInit(plist, 2);


	partPush(plist[0], &ps->s_iphost, sizeof(IPhost));
	partPush(plist[0], &s_port, sizeof(long));


	if (spoofmode) {
	    IPhost spoofaddr;

	    xTrace1(proxyrlp, TR_EVENTS, "PROXYRL spoofing address %s", pcspoofaddr);
	    str2ipHost(&spoofaddr, pcspoofaddr);
	    partPush(plist[1], &spoofaddr, sizeof(IPhost));
	    partPush(plist[1], &c_port, sizeof(long));
	}

	if ( ps->spi != 0 )
	{
		partPush(plist[2], &(ps->spi), sizeof(XSaid));
 		partPush(plist[1], &(ps->myhost), sizeof(IPhost));
 		partPush(plist[1], &c_port, sizeof(long));
	}
	else if ( ps->algorithm != NULL )
	{
		XAlgorithm algo;
		XObj lls;
    		Part pi[3];
		IPhost myhost;
   		algo = xControl(xGetDown(self,1),SAIDMGR_RESOLVE_NAME,
			ps->algorithm,strlen(ps->algorithm));
		if ( algo < 0 ) 
 			Kabort("client_init: PROXYRL unknown security algorithm specified");
 	    	xTrace1(proxyrlp, TR_EVENTS, "ROXYRL algorithm no %o", algo);
   		partInit(pi,3);
   		partPush(pi[REMOTE_PART], &ps->s_iphost, sizeof(IPhost));
   		partPush(pi[LOCAL_PART], ANY_HOST, 0);
    		partPush(pi[2], &ps->mode, sizeof(XIpsecMode));
    		partPush(pi[2], &algo, sizeof(XAlgorithm));
   		lls = xOpen(self, self, xGetDown(self,1),pi);
		if ( lls == ERR_XOBJ ) 
 			Kabort("client_init: PROXYRL couldn't open saidmgr session");
 		said=xControl(lls,SAIDMGR_GETLSAID,NULL,0);
 		xControl(lls,GETMYHOST,(char *)&myhost,sizeof(IPhost));
 		partPush(plist[1], &myhost, sizeof(IPhost));
 		partPush(plist[1], &c_port, sizeof(long));
	    	xTrace1(proxyrlp, TR_EVENTS, "ROXYRL said no %u", said);
		partPush(plist[2], &said, sizeof(XSaid));
	}

	partPush(plist[1], ps->locuser, strlen(ps->locuser)+1);
	partPush(plist[1], ps->remuser, strlen(ps->remuser)+1);
	partPush(plist[1], ps->termtype, strlen(ps->termtype)+1);

	fprintf(stderr, "PROXYRL client_init: opening session %d...\n", i);

	if ((prlsession = xOpen(self, self, self, plist)) == ERR_XOBJ) {
	    fprintf(stderr, "client_init: xOpen failed...\n");
	    xAssert(0);
	}

	if (vtapmode) {
	    xControl(xGetDown(prlsession, 0), VTAP_ENABLETAP, 0, 0);
	    xControl(xGetDown(prlsession, 0), VTAP_PRINTCHARS, 0, 0);
	} else {
	    xControl(xGetDown(prlsession, 0), VTAP_DISABLETAP, 0, 0);
	}

	if (prlsession == ERR_XOBJ) {
	    xTrace0(proxyrlp, TR_GROSS_EVENTS, "PROXYRL client_init: couldn't get lls");
	} else {
	    xTrace0(proxyrlp, TR_GROSS_EVENTS, "PROXYRL client_init: finished xOpen");
	}

	ss = prlsession->state;
	ss->index = i;

	if (i == -1) {
	    evDetach(evSchedule(client_kbdpoll, prlsession, 0));
	} else {
	    FILE *fpin, *fpout;

	    if (!(fpin = fopen(pscripts[i].infile, "r"))) {
		fprintf(stderr, "PROXYRL fopen(): %s", pscripts[i].infile);
		perror(NULL);
		xAssert(0);
	    }
	    if (!(fpout = fopen(pscripts[i].outfile, "w"))) {
		fprintf(stderr, "PROXYRL fopen(): %s", pscripts[i].outfile);
		perror(NULL);
		xAssert(0);
	    }
	    ss->fpin = fpin;
	    ss->fpout = fpout;

	    evDetach(evSchedule(client_scripthandle, prlsession, 0));
	}
    }
}

void 
client_kbdpoll(Event e, void *tself)
{
    fd_set pollfds;
    XObj self, lls;
    int numret;
    Msg msg;
    char c;
    struct timeval timeout = {0, 0};

    xTrace0(proxyrlp, TR_MORE_EVENTS, "PROXYRL Entering client_kbdpoll");

    self = (XObj) tself;
    if (!xIsSession(lls = xGetDown(self, 0))) {
	Kabort("PROXYRL could not get lower protocol");
    }

    /* we want non-canonical keyboard control */
    system("stty cbreak");

    /*
    if (!xIsSession(lls = xGetDown(self, 0))) {
	Kabort("client_kbdpoll: PROXYRL could not get lower session");
    }
    */

    for (;;) {
	FD_ZERO(&pollfds);
	FD_SET(0, &pollfds);

	if (select(1, &pollfds, (fd_set *) 0, (fd_set *) 0, &timeout) < 0) {
	    perror("client_kbdpoll: select()");
	    xAssert(0);
	}

	if (!FD_ISSET(0, &pollfds)) {
	    Delay(5);
	    continue;
	}

	if ((numret = read(0, &c, 1)) != 1) {
	    fprintf(stderr, "client_kbdpoll: read error!  returned %d\n",
	    		numret);
	    xAssert(0);
	}

	msgConstructBuffer(&msg, &c, 1);
	xPush(lls, &msg);
	msgDestroy(&msg);

	xControl(lls, TCP_PUSH, 0, 0);
    }

    xTrace0(proxyrlp, TR_MORE_EVENTS, "PROXYRL Exiting client_kbdpoll");
}

void 
client_scripthandle(Event e, void *tself)
{
    XObj self, lls;
    Msg msg;
    int i;
    char buf[1024];
    SState *ss;

    xTrace0(proxyrlp, TR_MORE_EVENTS, "PROXYRL Entering client_scripthandle");

    self = (XObj) tself;
    if (!xIsSession(lls = xGetDown(self, 0))) {
	Kabort("PROXYRL could not get lower protocol");
    }
    ss = self->state;

    while (fgets(buf, sizeof(buf), ss->fpin) != NULL) {

	/* is it a special command? */
	if (strncmp(buf, "@sleep", 6) == 0) {
	    int numsecs, err;
	    char trash[1024];

	    if ((err = sscanf(buf, "%s %d", trash, &numsecs)) != 2) {
	        fprintf(stderr, "%s: parse error (sscanf returned %d)...\n",
		    pscripts[i].infile, err);
		xAssert(0);
	    }

	    Delay(numsecs * 100);
	}

	msgConstructBuffer(&msg, buf, strlen(buf));
	xPush(lls, &msg);
	msgDestroy(&msg);

	xControl(lls, TCP_PUSH, 0, 0);
    }

    xTrace0(proxyrlp, TR_MORE_EVENTS, "PROXYRL Exiting client_scripthandle");
}

#define TCPBUFSIZ 4096

static xkern_return_t 
proxyrlDemux(XObj self, XObj lls, Msg * msg)
{
    PState *pstate = self->state;
    XObj prls;

    xTrace0(proxyrlp, TR_MORE_EVENTS, "PROXYRL Entering clientdemux");

    /* we have to resolve first */
    if (mapResolve(pstate->activeMap, &lls, &prls) == XK_FAILURE) {
	xTrace0(proxyrlp, TR_MAJOR_EVENTS, "PROXYRL Demux: bad mapResolve()");
	return XK_FAILURE;
    }

    return proxyrlPop(prls, lls, msg, 0);
}

static xkern_return_t
proxyrlPop(s, lls, msg, arg)
    XObj s;
    XObj lls;
    Msg *msg;
    VOID *arg;
{
    int len;
    static char buf[8192];
    u_short bufsize = TCPBUFSIZ;
    SState *ss;
    int sessionnumber;

    xTrace1(proxyrlp, TR_EVENTS, "PROXYRL pop, length = %d", msgLen(msg));
    xAssert(xIsSession(s));

    ss = s->state;

    len = msgLen(msg);
    xAssert(len < sizeof(buf));

    sessionnumber = ss->index;
    /*fprintf(stderr, "clientdemux (session %d) : %d pending...\n",
			    sessionnumber, len);*/

    msgPop(msg, mycopypop, buf, len, 0);
    buf[len] = '\0';		/* terminate the buffer */

    /* if the session number is -1, then it's the interactive tty.
     * otherwise, it's being driven by a script.  here, we writen not
     * to the terminal, but the a predetermined output stdio descriptor
     */
    /* if it's a tty */
    if (sessionnumber < 0) {
	fputs(buf, stdout);
	fflush(stdout);
    }
    /* otherwise, it's a file */
    else {
	fputs(buf, ss->fpout);
	fflush(ss->fpout);
    }


    xAssert(xControl(lls, TCP_SETRCVBUFSPACE, (char *) &bufsize, sizeof(bufsize))
    		== XK_SUCCESS);

    xTrace0(proxyrlp, TR_MORE_EVENTS, "PROXYRL Exiting clientdemux");

    return XK_SUCCESS;
}

static int
proxyrlControlProtl(self, opcode, buf, len)
    XObj self;
    int opcode, len;
    char *buf;
{
    /*
     * All opcodes are forwarded to the lower protocol
     */
    return xControl(xGetDown(self, 0), opcode, buf, len);
}

static int
proxyrlControlSessn(self, opcode, buf, len)
    XObj self;
    int opcode, len;
    char *buf;
{
    /*
     * All opcodes are forwarded to the lower protocol
     */
    return xControl(xGetDown(self, 0), opcode, buf, len);
}

static void
sessnInit(s)
    XObj s;
{
    xAssert(xIsSession(s));


    s->control = proxyrlControlSessn;
    s->pop = proxyrlPop;
#ifdef 0
    s->close = udpCloseProtl;
    s->open = udpOpen;
    s->openenable = udpOpenEnable;
    s->opendisable = udpOpenDisable;
    s->demux = udpDemux;
    s->opendisableall = udpOpenDisableAll;
    s->push = vtapPush;
    s->close = vtapClose;
    s->control = vtapControlSessn;
    /*
     * VTAP sessions will look like a protocol to lower sessions, so we
     * need a demux function
     */
    s->demux = vtapSessnDemux;
#endif
}

static void
protInit(p)
    XObj p;
{
    xAssert(xIsProtocol(p));

    p->demux = proxyrlDemux;
    p->open = proxyrlOpen;
    p->control = proxyrlControlProtl;
    /*
    p->opendisable = vtapOpenDisable;
    p->demux = vtapProtlDemux;
    p->opendone = vtapOpenDone;
    */
}

