/*#include <sys/systeminfo.h>*/
#include <sys/types.h>
#include <time.h>
#include <stdlib.h>
#include "xkernel.h"
#include "tcp.h"
#include "ip.h"
#include "udp.h"
#include "x_stdio.h"
#include "rc_internal.h"
#include "rc.h"

void rc_init (self)
     Protl self;
{
  PState        *ps;
  
  getproc_protl (self);
  
  ps = X_NEW(PState);
  bzero((char *)ps, sizeof(PState));
  self->state = (VOID *)ps;
  ps->RC_sessn = NULL;

  if (!xIsProtocol(xGetDown(self, 0))) 
    Kabort("RC could not get lower protocol");
} 


static Sessn
rcOpen( self, hlp, hlpType, p )
    Protl self, hlp, hlpType;
    Part *p;
{
  Sessn lls;
  PState *ps = (PState *)self->state;
  
  if (ps->RC_sessn == NULL)
    {
      lls = xOpen(self, self, xGetDown(self, 0), p);
      if (lls == ERR_SESSN)
	{
	  Kabort ("RC could not open llp!");
	}
      ps->lls = lls;
      ps->RC_sessn = rcCreateSessn (hlp, hlpType, self, lls);
   }
  else
    {
      xTrace0 (rcp, TR_ERRORS, 
	       "rcOpen: hlp tries to re-open RC protocol");
      return ERR_SESSN;
    }
  
  /* fire off the rate process */
  
  return ps->RC_sessn;
}


static Sessn
rcCreateSessn(self, hlp, hlpType, lls)
    Protl self, hlp, hlpType;
    Sessn lls;
{
  Sessn  s = ERR_SESSN;
  SState *ss;
  
  s = xCreateSessn(getproc_sessn, hlp, hlpType, self, 
		   1, &lls);
  if (s == ERR_SESSN)
    {
      Kabort ("Could not create RC session!");
    }
  
  /* create session state */
  ss = X_NEW(SState);
  s->state = (VOID *)ss;
  bzero((char *)ss, sizeof(SState));

  ss->MsgsToSend = MAX_MSGS;
  ss->BytesToSend = MAX_BYTES;
  semInit (&ss->Sem, 0);
  semInit (&ss->queueSem, 0);
  ss->MsgQueue.next = ss->MsgQueue.prev = &(ss->MsgQueue);

  ss->rateProcess = evSchedule (rcRateController, (void *)ss, 0);
  ss->queueMgr = evSchedule (rcQueueManager, (void *) s, 0);
  
  return s;
}


static int
rcControlSessn(s, opcode, buf, len)
    Sessn s;
    int opcode;
    char *buf;
    int len;
{
  return xControl(xGetDown (s, 0), opcode, buf, len);
}


static int
rcControlProtl(self, opcode, buf, len)
    Protl self;
    int opcode;
    char *buf;
    int len;
{
  return xControl(xGetDown (self, 0), opcode, buf, len);
}



static xkern_return_t
rcOpenEnable( self, hlp, hlpType, p )
    Protl self, hlp, hlpType;
    Part *p;
{
  /* should never be called */
  return XK_SUCCESS;
}


static xkern_return_t
rcClose(self)
    Sessn self;
{
  return xClose (xGetDown (self, 0));
}


static xmsg_handle_t
rcPush(s, msg)
    Sessn s;
    Msg *msg;
{
  SState *ss = (SState *) s->state;
  RCinfo  i;
  
  if (MsgQueued (ss))
    {
      rcEnqueueMsg (ss, msg);
      if ((! ss->queueMgrRunning) && MsgCanGo (ss, ss->MsgQueue.next->m))
	{
	  ss->queueMgrRunning = 1;
	  semSignal (&ss->queueSem);
	}
    }
  else
    {
      if (MsgCanGo (ss, msg))
	{
	  xTrace0 (rcp, TR_GROSS_EVENTS, "rcPush: message proceeds");
	  i = (RCinfo) msgGetAttr (msg, 0);
	  if (i != NULL)
	    rcInternalPush (s, msg, i->tcp_s, i->len);
	  else
	    rcInternalPush (s, msg, NULL, 0);
	}
      else
	rcEnqueueMsg (ss, msg);
    }

  return XK_SUCCESS;
}


static xkern_return_t
rcDemux(self, lls, dg)
    Protl self;
    Sessn lls;
    Msg *dg;
{
  /* should never be called */
  return XK_SUCCESS;
}


static xkern_return_t
rcPop(s, ds, dg, arg)
    Sessn s;
    Sessn ds;
    Msg *dg;
    VOID *arg;
{
  /* should never be called */
  return XK_SUCCESS;
}


static void
getproc_protl(p)
    Protl p;
{
  p->control = rcControlProtl;
  p->open = rcOpen;
  p->close = rcClose;
  p->openenable = rcOpenEnable;
  p->demux = rcDemux;
}
 

static void
getproc_sessn(s)
    Sessn s;
{
  s->push = rcPush;
  s->control = rcControlSessn;
  s->close = rcClose;
  s->pop = rcPop;
}
 

static void rcRateController (Event e, void *arg)
{
  SState *ss = (SState *) arg;
  int i;

  xTrace0 (rcp, TR_GROSS_EVENTS, "rcRateController active!");
  i = 1;
  while (ss->rateProcess != NULL)
    {
      Delay (TIME_UNIT / SUB_INT);
      ss->MsgsToSend += ss->Info[i].MsgsSent;
      ss->BytesToSend += ss->Info[i].BytesSent;
      ss->Info[i].MsgsSent = ss->Info[i].BytesSent = 0;
      ss->current = (ss->current + 1) % SUB_INT;
      i = (i + 1) % SUB_INT;

      if ((! ss->queueMgrRunning) && MsgQueued (ss) && 
	  MsgCanGo (ss, ss->MsgQueue.next->m))
	{
	  ss->queueMgrRunning = 1;
	  semSignal (&ss->queueSem);
	}
      
    }
}


static void rcQueueManager ( Event e, void *arg )
{
  Sessn s = (Sessn) arg;
  SState *ss = (SState *) s->state;
  Msg *m;
  Sessn tcp_s;
  u_short len;

  xTrace0 (rcp, TR_GROSS_EVENTS, "rcQueueManager active!");
  
  while (ss->queueMgr != NULL)
    {
      xTrace0 (rcp, TR_GROSS_EVENTS, "rcQueueMananger sleeps...");
      ss->queueMgrRunning = 0;
      semWait (&ss->queueSem);
      xTrace0 (rcp, TR_GROSS_EVENTS, "rcQueueMananger awakes...");
      while (MsgQueued (ss) && MsgCanGo (ss, ss->MsgQueue.next->m))
	{
	  m = rcDequeueMsg (ss, &tcp_s, &len);
	  rcInternalPush (s, m, tcp_s, len);
	  msgDestroy (m);
	  xFree ((char *) m);
	}
    }
  
}


static void rcEnqueueMsg ( SState *s, Msg *msg )
{
  queueNode *n = X_NEW (queueNode);
  Msg *new = X_NEW(Msg);
  RCinfo    i;
  
  n->next = &s->MsgQueue;
  n->prev = s->MsgQueue.prev;
  s->MsgQueue.prev = s->MsgQueue.prev->next = n;
  
  msgConstructCopy (new, msg);
  n->m = new;

  n->num = s->queuenum++;
  xTrace1 (rcp, TR_MAJOR_EVENTS, "Enqueued %d", n->num);
  
  /* 
   * Right now, the RCinfo struct lives in the stack frame of TCPF --
   * so no garbage collection is needed.
   */
  i = (RCinfo) msgGetAttr (msg, 0);
  if (i != NULL)
    {
      n->tcp_s = i->tcp_s;
      n->len = i->len;
    }
  else
    {
      n->tcp_s = NULL;
      n->len = 0;
    }
}


static Msg *rcDequeueMsg ( SState *s, Sessn *tcp_s, u_short *len )
{
  queueNode *n;
  Msg *m;
  
  if (s->MsgQueue.next == &s->MsgQueue)
    return NULL;
  
  n = s->MsgQueue.next;
  n->next->prev = &s->MsgQueue;
  s->MsgQueue.next = n->next;

  xTrace1 (rcp, TR_MAJOR_EVENTS, "Dequeued %d", n->num);
  
  *tcp_s = n->tcp_s;
  *len = n->len;
  m = n->m;
  xFree ((char *) n);
  
  return m;
}


static void rcInternalPush (Sessn s, Msg *msg, Sessn tcp_s, u_short len)
{
  SState *ss = (SState *) s->state;
  int m = msgLen(msg), c = ss->current;

  ss->MsgsToSend --;
  ss->BytesToSend -= m;
  ss->Info[c].MsgsSent++;
  ss->Info[c].BytesSent += m;

  if (xPush (xGetDown (s, 0), msg) != XK_SUCCESS)
      xTrace0 (rcp, TR_ERRORS, "rcInternalPush:  push failed!");
 
  xTrace2 (rcp, TR_GROSS_EVENTS, 
	   "rcInternalPush: MsgsToSend = %d, BytesToSend = %d",
	   ss->MsgsToSend, ss->BytesToSend);

  /* 
   * Note: The semantics of TCP's control-op have been changed to
   *       meet the needs of the TCP Forwarder.  It increments the
   *       window size by 'len'.
   */
  if (tcp_s != NULL)
    {
      if (xControl(tcp_s, TCP_SETRCVBUFSPACE, (char *)&len, sizeof(len)) < 0)
	xError("rcInternalPush: TCP_SETRCVBUFSPACE failed");

      /* Decrement ref cnt of TCP session (previously incremented by TCPF) */
      xClose (tcp_s);
    }
}
