/* sendauth.c
 *
 * this protocol should emulate the sendauth/recvauth functions
 * of kerberos.  They can be found in krb-src/lib/krb/sendauth.c
 *
 */


#include "xkernel.h"
#include "prot/sendauth.h"
#include "sendauth_i.h"
#include <krb5/krb5.h>
#include <isode/psap.h>
#include <krb5/kdb.h>
#include <krb5/asn1.h>
#include <com_err.h>
#include "prot/join.h"
#include "prot/crypt.h"


#ifdef __STDC__

xkern_return_t sendauthDemux(XObj, XObj, Msg *);
XObj           sendauthOpen(XObj, XObj, XObj, Part *);
int            sendauthProtControl(XObj, int, char *, int);
int            sendauthSessnControl(XObj, int, char *, int);
static void    mycopypush(void *, char *, long, void *);
static void    sessn_init( XObj );
xmsg_handle_t  sendauthPush( XObj, Msg * );
xkern_return_t sendauthPop( XObj, XObj, Msg *, void * );
static long    mycopypop(void *, char *, long, void *);


#else

xkern_return_t sendauthDemux();
XObj           sendauthOpen();
int            sendauthProtControl();
int            sendauthSessnControl();
static void    mycopypush();
static void    sessn_init();
xmsg_handle_t  sendauthPush();
xkern_return_t sendauthPop();
static long    mycopypop();


#endif


int tracesendauthp = 1;




char *sendauth_version = "KRB5_SENDAUTH_V1.0";


void sendauth_init(self)
     XObj self;
{
  self->state = (void *)xMalloc(sizeof(PSTATE));
  bzero((char *)(self->state), sizeof(PSTATE));

  ((PSTATE*)self->state)->activemap = mapCreate(20, sizeof(XObj));

  self->open = sendauthOpen;
  self->demux = sendauthDemux;
  self->control = sendauthProtControl;

}



XObj sendauthOpen(self, hlpType, hlpRcv, part)
     XObj self, hlpType, hlpRcv;
     Part *part;
{
  PSTATE         *ps = (PSTATE *)self->state;
  SSTATE         *sstate;
  SendauthArgs   *args;
  Msg             msg;
  int             len, retval, i;
  XObj            s;
  char            partBuf[256];
  Msg             msgs[4];
  char            ck_buf[16];
  krb5_ap_req     ap_req;
  


  xTrace0(sendauthp, TR_MAJOR_EVENTS, "sendauthOpen -- entered");

  if (partLen(part) < 1) {
    xTrace0(sendauthp, TR_ERRORS, "bad participant list");
    return ERR_XOBJ;
  }
  args = (SendauthArgs *)partPop(*part);


  s = xCreateSessn(sessn_init, hlpRcv, hlpType, self, 0, 0);
  sstate = (SSTATE *)xMalloc(sizeof(SSTATE));
  if (!sstate) {
    xTrace0(sendauthp, TR_ERRORS, "malloc failed for sstate");
    return ERR_XOBJ;
  }
  s->state = (void *)sstate;
  bzero((char*)sstate, sizeof(SSTATE));
  sstate->args = args;
  semInit(&sstate->wait, 0);

  sstate->lower[0] = xOpen(self, self, xGetDown(self, 0), part);
  if (sstate->lower[0] == ERR_XOBJ) {
    xTrace0(sendauthp, TR_ERRORS,
	    "Could not open sendauth_bottom control session");
    return ERR_XOBJ;
  }
  
  if (xControl(sstate->lower[0], GETPARTICIPANTS, partBuf, 256) <= 0) {
    xTrace0(sendauthp, TR_ERRORS, "could not get participants");
    return ERR_XOBJ;
  }


  mapBind(ps->activemap, &(sstate->lower[0]), s);
  if (xControl(sstate->lower[0], JOINSETCONTROL, (char *)&s, sizeof(XObj)) <= 0) {
    xTrace0(sendauthp, TR_ERRORS, "Could not set control session for sendauth");
    return ERR_XOBJ;
  }

  for(i=1; i<7;i++) {
    partInternalize(part, partBuf);
    sstate->lower[i] = xOpen(self, self, xGetDown(self,i), part);
    if (sstate->lower[i] == ERR_XOBJ) {
      xTrace1(sendauthp, TR_ERRORS, "could not open lls %d", i);
      return ERR_XOBJ;
    }
    if (i == 2 || i == 4) {
      if (xControl(sstate->lower[i], JOINSETCONTROL, (char *)&s, sizeof(XObj)) <= 0) {
	xTrace1(sendauthp, TR_ERRORS, "Could not set control session for lls %d", i);
	return ERR_XOBJ;
      }
    }

    mapBind(ps->activemap, &sstate->lower[i], s);
  }


       /* need to make 4 messages */
  /* message to checksum ===> lower[6] */
  /*              client ===> lower[5] */
  /*              ap_req ===> lower[3] */
  /*    (SendauthArgs *) ===> lower[1] */


  bzero(ck_buf, 16);

  ap_req.ap_options = 0;
  if (retval = decode_krb5_ticket(&args->credsp->ticket, &ap_req.ticket)) {
    xTrace1(sendauthp, TR_ERRORS,
	    "sendauthOpen -- error (%d) while decoding ticket", retval);
    return ERR_XOBJ;
  }

  if (xControl(sstate->lower[4], CRYPT_OUT_SET,
	       (char*)args->credsp->keyblock.contents,
	       args->credsp->keyblock.length) != 8) {
    xTrace0(sendauthp, TR_ERRORS, "Error -- CRYPT_OUT_SET failed!");
    return ERR_XOBJ;
  }

  msgConstructBuffer(&msgs[0], ck_buf, 16);
  msgConstructBuffer(&msgs[1], (char*)&args->client, sizeof(krb5_principal));
  msgConstructBuffer(&msgs[2], (char*)&ap_req, sizeof(krb5_ap_req));
  msgConstructBuffer(&msgs[3], (char*)&args, sizeof(SendauthArgs *));


  xPush(sstate->lower[6], &msgs[0]);
  xPush(sstate->lower[5], &msgs[1]);
  xControl(sstate->lower[4], JOINDONE, 0, 0);
  
  xPush(sstate->lower[3], &msgs[2]);
  xControl(sstate->lower[2], JOINDONE, 0, 0);

  xPush(sstate->lower[1], &msgs[3]);
  xControl(sstate->lower[0], JOINDONE, 0, 0);

 semWait(&sstate->wait);

  if (sstate->status == XK_SUCCESS) {
    xTrace0(sendauthp, TR_DETAILED, "SendauthOpen -- sendauth successful");
    return s;
  } else {
    xTrace0(sendauthp, TR_DETAILED, "SendauthOpen -- sendauth not successful");
    return ERR_XOBJ;
  }
}


static void sessn_init(self)
     XObj self;
{
  self->pop = sendauthPop;
  self->control = sendauthSessnControl;
  self->push = sendauthPush;
}


xkern_return_t sendauthDemux(self, lls, msg)
     XObj self, lls;
     Msg *msg;
{
  PSTATE *ps = (PSTATE *)self->state;
  XObj    s;

  xTrace0(sendauthp, TR_EVENTS, "sendauthDemux -- entered");

  if (mapResolve(ps->activemap, &lls, &s) == XK_SUCCESS) {
    return xPop(s, lls, msg, 0);
  } else {
    xTrace1(sendauthp, TR_ERRORS, "no binding found for lls 0x%x", lls);
    return XK_FAILURE;
  }
}


xmsg_handle_t sendauthPush(s, msg)
     XObj  s;
     Msg  *msg;
{
  xTrace0(sendauthp, TR_ERRORS,
	  "Error -- no messages should be passed through sendauth!");
  return XMSG_ERR_HANDLE;
}


xkern_return_t sendauthPop(s, lls, msg, hdr)
     XObj s, lls;
     Msg *msg;
     void *hdr;
{
  char     one_byte = 1;
  SSTATE  *ss = (SSTATE *)s->state;


  /* read one byte from the message.  If the byte equals zero,
     then the sendauth protocol was successful.  Otherwise,
     return failure;
  */

  xTrace0(sendauthp, TR_MAJOR_EVENTS, "SendauthPop -- entered");

  if (msgLen(msg) >= 1) {
    msgPop(msg, mycopypop, &one_byte, 1, 0);
    if (one_byte == 0) {
      xTrace0(sendauthp, TR_EVENTS, "SendauthPop -- successful transaction");
      ss->status = XK_SUCCESS;
      semSignal(&ss->wait);
      return XK_SUCCESS;
    }
  }

  /* if we are here then something went wrong */

  xTrace0(sendauthp, TR_EVENTS, "SendauthPop -- transaction not successful");
  ss->status = XK_FAILURE;
  semSignal(&ss->wait);
  return XK_FAILURE;
}
    
int sendauthProtControl(self, op, buf, len)
     XObj self;
     int op, len;
     char *buf;
{
  return xControl(xGetDown(self, 0), op, buf, len);
}

int sendauthSessnControl(self, op, buf, len)
     XObj self;
     int op, len;
     char *buf;
{
  return xControl(xGetDown(self, 0), op, buf, len);
}

  

static void mycopypush(src, dst, count, arg)
     void *src;
     char *dst;
     long  count;
     void *arg;
{
  bcopy((char *)src, dst, count);
}

static long mycopypop(dest, src, count, arg)
     void *dest;
     char *src;
     long count;
     void *arg;
{
  bcopy(src, (char*)dest, count);
  return count;
}
