#include "xkernel.h"
#include "tcp.h"
#include "telnet_i.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <pwd.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#ifdef MACH
#ifdef INTELX86
#include <sysent.h>
#endif
#endif

#ifdef SUNOS
#include <unistd.h>
#include <stdlib.h>
#endif

#ifdef SOLARIS
#include <unistd.h>
#include <sys/termios.h>
#include <sys/ttold.h>
#include <stdlib.h>
#undef HUPCL
#undef XTABS
#undef ECHO
#undef CR1
#undef CR2
#undef CR3
#undef NL1
#undef TAB1
#undef FF1
#undef BS1
#include <sgtty.h>
#ifdef CLTECH
#undef CLTECH
#endif
#define CTLECH 0x10000000      /* echo control */
#ifdef CBREAK
#undef CBREAK
#endif
#define CBREAK 0x00000002      /* half-cooked mode */ 
#endif
#ifdef ECHO
#undef ECHO
#endif
#define ECHO 0x00000008      /* echo input */
#ifdef RAW
#undef RAW
#endif
#define RAW 0x00000020  /* no i/o processing */

#ifdef IRIX
#include <sys/ttold.h>
#ifdef CLTECH
#undef CLTECH
#endif
#define CTLECH 0x10000000      /* echo control */
#ifdef CBREAK
#undef CBREAK
#endif
#define CBREAK 0x00000002      /* half-cooked mode */ 
#endif


int	tracetelnetp;


static XObj		telnetCreateSessn( XObj, XObj, XObj, ActiveId * );
static xkern_return_t	telnetDemux( XObj, XObj, Msg * );
static xkern_return_t	telnetCloseSessn( XObj );
static xkern_return_t	telnetCloseDone( XObj );
static void		getproc_sessn(XObj);
void			telnet_init(XObj);
static bool		write_message(char *, long, void *);
static void 		poll_it(Event, void *);

static	char		termEofChar;
static  char		termEraseChar;

static void poll_it(ev, me)
Event ev; 
void * me;
{
	XObj	self = (XObj) me;
	SSTATE * ss  = (SSTATE *)self->state; 
	PSTATE * ps  = (PSTATE *)self->myprotl->state; 

	fd_set	readfdset;
    	struct timeval timeout;
	int	nfds = 64;
	unsigned char	rbuf;
	Msg	msg;

	timeout.tv_sec = 0;
	timeout.tv_usec = 0;

	for(;;)
	{
		FD_ZERO(&readfdset);
		FD_SET(ss->input_handle, &readfdset);

		if (select(nfds, &readfdset, (fd_set *) 0, (fd_set *) 0, &timeout) < 0) {
	    		Kabort("telnet_poll: select() failed!");
		}

		if (!FD_ISSET(ss->input_handle,&readfdset)) 
		{
			
			evDetach( ps->poll = evSchedule(poll_it,self, POLL_TIME));
			return;
		}
		else if  (read(ss->input_handle,(char *)&rbuf,1) != 1)
		{
			/* got CTRL-D */
			
			rbuf = termEofChar; /*EOF*/
			xTrace1(telnetp, TR_EVENTS,
				"telnet_init: pushing char <%03u>",rbuf);
		     	msgConstructBuffer(&msg,(char *)&rbuf,1);
		    	xPush(xGetDown(self,0),&msg);
			msgDestroy(&msg);
			evDetach( ps->poll = evSchedule(poll_it,self, POLL_TIME));
			return;
		}
		else
		{
			switch(rbuf)
			{
				case 3:
					xClose(xGetDown(self,0));
					exit(0);
					break;
				case 13:
					rbuf = 13;
					xTrace1(telnetp, TR_EVENTS,
						"telnet_init: pushing char <%03u>",rbuf);
		     			msgConstructBuffer(&msg,(char *)&rbuf,1);
		    			xPush(xGetDown(self,0),&msg);
					msgDestroy(&msg);
					rbuf = 10;
					xTrace1(telnetp, TR_EVENTS,
						"telnet_init: pushing char <%03u>",rbuf);
		     			msgConstructBuffer(&msg,(char *)&rbuf,1);
		    			xPush(xGetDown(self,0),&msg);
					msgDestroy(&msg);
					break;		
				default:
					xTrace1(telnetp, TR_EVENTS,
						"telnet_init: pushing char <%03u>",rbuf);
		     			msgConstructBuffer(&msg,(char *)&rbuf,1);
		    			xPush(xGetDown(self,0),&msg);
					msgDestroy(&msg);
			}
		}
	}
}

static bool write_message(Ptr, len, ss)
char * Ptr;
long len;
void * ss;
{
    int x;
    XObj self = (XObj) ss;
    SSTATE * s = (SSTATE *) self->state;
    unsigned char rbuf[3];
    Msg	msg;

    for ( x = 0; x < len; x++)
    {
	xTrace1(telnetp, TR_EVENTS,
		"telnet_init: poping char <%03u>",(unsigned char) Ptr[x]);
	switch (s->expecting_command)
	{
		case 0:
		switch ((unsigned char)Ptr[x])
		{
			case 0:
				break;
			case 7: /* bell */
			case 9: /* HT */
			case 11: /*VT */
			case 12: /*FF */
				break;
			case 8: /* BS */
				write(s->output_handle, &termEraseChar, 1);
				break;
			case 255:
				s->expecting_command = 1;
				break;
			default:
				write(s->output_handle, &Ptr[x], 1);
		}
		break;
		case 1:
		switch ((unsigned char)Ptr[x])
		{
			case 253: /* Do ? */
				s->expecting_command = 2;
				break;
			case 251: /*Will ? */
				s->expecting_command = 3;
				break;
			case 252: /* Won't */
				s->expecting_command = 5;
				break;
			case 254: /* Don't */
				s->expecting_command = 6;
				break;
			case 255:
				write(s->output_handle, &Ptr[x], 1);
				break;
			default:
				s->expecting_command = 0;
		}
		break;
		case 2: /*DO ? */
		switch ((unsigned char)Ptr[x])
		{
			default:
				rbuf[0]=255;
				rbuf[1]=252;	/* Won't */
				rbuf[2]=(unsigned char)Ptr[x];
				xTrace3(telnetp, TR_EVENTS,
					"telnet_init: pushing char <%03u,%03u,%03u>",
					rbuf[0],rbuf[1],rbuf[2]);
		     		msgConstructBuffer(&msg,rbuf,3);
		    		xPush(xGetDown(self,0),&msg);
				msgDestroy(&msg);
				s->expecting_command = 0;
		}
		break;
		case 3: /*WILL ? */
		switch ((unsigned char)Ptr[x])
		{
			default: 
				rbuf[0]=255;
				rbuf[1]=254;	/* Don't */
				rbuf[2]=(unsigned char)Ptr[x];
				xTrace3(telnetp, TR_EVENTS,
					"telnet_init: pushing char <%03u,%03u,%03u>",
					rbuf[0],rbuf[1],rbuf[3]);
		     		msgConstructBuffer(&msg,rbuf,3);
		    		xPush(xGetDown(self,0),&msg);
				msgDestroy(&msg);
				s->expecting_command = 0;
		}
		break;
		case 5: /* Got Won't  */
		switch ((unsigned char)Ptr[x])
		{
			default:
				s->expecting_command = 0;
		}
		break;
		case 6: /*Got  Don't  */
		switch ((unsigned char)Ptr[x])
		{
			case 01:  /* Echo on server*/
				{
				struct sgttyb sb;
    				ioctl(s->input_handle,TIOCGETP,(char *) &sb);
    				sb.sg_flags &= ~ECHO;
    				sb.sg_flags &= ~CTLECH;
    				ioctl(s->input_handle,TIOCSETP,(char *) &sb);
				s->expecting_command = 0;
				}
				break;
			default:
				s->expecting_command = 0;
		}
		break;
		default:
				s->expecting_command = 0;
	}

    }
    return TRUE;
}


/*
 * telnet_init
 */
void
telnet_init(self)
    XObj self;
{
    PSTATE	*pstate;
    ActiveId 	key;
    Part	part[3];
    XObj	llp;
    XObj	ses;
    long	localPort;
    long	remotePort = 23;
    XTime	t;

    xTrace0(telnetp, TR_GROSS_EVENTS, "TELNET init");
    xGetTime(&t);
    srand(t.sec);
    localPort = (rand() % 50000) + 4000;
    pstate = X_NEW(PSTATE);
    bzero((char *)pstate, sizeof(PSTATE));
    self->state = (VOID *)pstate;
    self->demux = telnetDemux;
    self->closedone = telnetCloseDone;
    if (globalArgc != 2)
	Kabort("TELNET illegal number of arguments :: telnet <IP address>");
    str2ipHost((IPhost*) &(pstate->telnet_to), globalArgv[1]);
    if ( ! xIsProtocol(llp = xGetDown(self, 0)) ) {
	Kabort("TELNET could not get lower protocol");
    }
    xControl(llp,GETMYHOST,(char *)&(pstate->my_address), sizeof(IPhost));
    pstate->activemap = mapCreate(ACTIVE_MAP_SIZE, sizeof(ActiveId));
    partInit(part, 2);
    partPush(part[LOCAL_PART], &pstate->my_address, sizeof(IPhost));
    partPush(part[LOCAL_PART], &localPort, sizeof(long));
    partPush(part[REMOTE_PART], &pstate->telnet_to, sizeof(IPhost));
    partPush(part[REMOTE_PART], &remotePort, sizeof(long));
    if ( (key.lls = xOpen(self, self, llp, part)) == ERR_XOBJ ) {
	xTrace0(telnetp, TR_ALWAYS,
		"telnet_init: can't open transport protocol");
	xFree((char *) pstate);
	return;
    }
    ses = telnetCreateSessn(self,self,self,&key);
    evDetach(pstate->poll = evSchedule(poll_it,ses,POLL_TIME));

}

static XObj
telnetCreateSessn(self, hlpRcv, hlpType, key)
    XObj self, hlpRcv, hlpType;
    ActiveId *key;
{
    XObj	s;
    SSTATE	*sstate;
    PSTATE	*pstate;
    struct   tchars otc = { 0 };
    struct   sgttyb ottyb = { 0 };

    pstate = (PSTATE *)self->state;
    s = xCreateSessn(getproc_sessn, hlpRcv, hlpType, self, 1, &key->lls);
    s->binding = mapBind(pstate->activemap, key, s);
    sstate = X_NEW(SSTATE);
    s->state = sstate;
    sstate->input_handle = 0 ; /* stdin */
    sstate->output_handle = 1 ; /* stdout */ 
    sstate->expecting_command = 0;
    ioctl(sstate->input_handle, TIOCGETP, (char *)&ottyb);
    ioctl(sstate->input_handle, TIOCGETC, (char *)&otc);
    termEofChar = otc.t_eofc;
    termEraseChar = ottyb.sg_erase;
    ottyb.sg_flags &= ~CBREAK;
    ottyb.sg_flags |= RAW;
    ioctl(sstate->input_handle,TIOCSETP,(char *) &ottyb);
    return s;
}



/*
 * telnetDemux
 */
static xkern_return_t
telnetDemux(self, lls, dg)
    XObj self;
    XObj lls;
    Msg *dg;
{
    ActiveId  	activeid;
    PSTATE      *pstate;
    XObj	from_s;
    u_short	bufspace = 1024;

    pstate = (PSTATE *)self->state;
    xTrace0(telnetp, TR_EVENTS, "telnet Demux");
    activeid.lls = lls;
    if ( mapResolve(pstate->activemap, &activeid, &from_s) == XK_FAILURE ) 
    {
    	xTrace0(telnetp, TR_EVENTS, "telnet Demux: tcp open failed");
	return XK_FAILURE;
    }

    msgForEach(dg,write_message, from_s);
    xControl(lls, TCP_SETRCVBUFSPACE, (char *) &bufspace, sizeof(bufspace));	
    xTrace0(telnetp, TR_EVENTS, "telnet Demux: Done");
    return XK_SUCCESS;
}


/*
 * telnetCloseSessn
 */
static xkern_return_t
telnetCloseSessn(s)
    XObj s;
{
    SSTATE	*sstate;
    PSTATE	*pstate;

    sstate = (SSTATE *)s->state;
    pstate = (PSTATE *)s->myprotl->state;

 
    xAssert(xIsSession(s));
    xTrace1(telnetp, TR_MAJOR_EVENTS, "Forward close of session %x", s);
    xAssert(s->rcnt == 0);
    mapRemoveBinding(pstate->activemap,s->binding);
    xClose(xGetDown(s,0));
    xFree((char*)sstate);
    xDestroy(s);
    return XK_SUCCESS;
}
/*
 * telnetCloseSessn
 */
static xkern_return_t
telnetCloseDone(s)
    XObj s;
{
	exit(0);
}
static void
getproc_sessn(s)
    XObj s;
{
  xAssert(xIsSession(s));
  s->close = telnetCloseSessn;
}
