/*
 * $RCSfile: debug.c,v $
 *
 * Universal Stub Compiler
 *
 * Copyright (c) 1994    Arizona Board of Regents
 *
 * $Log: debug.c,v $
 * Revision 1.1  1994/10/13 17:43:30  dorgival
 * Initial revision
 *
 */

/*
 * Debug related functions: formatting, messages, etc.
 */

#include "usc.h"
#include "symbol.h"
#include "debug.h"
#include "codegen_export.h"
#include "y.tab.h"
#include "codegen.h"

char *bitorder_to_string (BitInfo *bits)
{
  static char *s = NULL;
  char tmp[128];
  int step, next, curr, range = 0, i;

  if (s == NULL)
    s = (char *)Malloc (MAX_BITFIELD_SIZE * sizeof (char));

  i = 0;

  strcpy (s, "<");
  while (bits[i] >= 0 && i < MAX_BITFIELD_SIZE) {
    curr = bits[i++];
    next = bits[i];
    sprintf(tmp,"%d",curr);
    strcat(s,tmp);
    if ( next >=0 ) {
      step = next - curr;
      if ( step == 1 || step == -1 ) {
	range = -1;
	do {
	  range++;
	  curr = next;
	  next = bits[++i];
	} while ( (next>=0)&&(next-curr == step) );
	if ( range )
	  sprintf(tmp,"..%d",curr);
	else
	  sprintf(tmp,",%d",curr);
	strcat(s,tmp);
      }
      if ( bits[i] >= 0 && i < MAX_BITFIELD_SIZE )
	strcat(s,",");
    }
  }
  strcat(s,">");

  return s;
}

char *byteorder_to_string (ByteOrder border)
{
  static char s[4*MAX_BORDER_SIZE];
  char tmp[128];
  int step,next,curr,range = 0;

  strcpy(s,"<");
  while (*border) {
    curr = (*border++)-1;
    next = (*border)-1;
    sprintf(tmp,"%d",curr);
    strcat(s,tmp);
    if ( next >=0 ) {
      step = next - curr;
      if ( step == 1 || step == -1 ) {
	range = -1;
	do {
	  range++;
	  curr = next;
	  next = (*(++border))-1;
	} while ( (next>=0)&&(next-curr == step) );
	if ( range )
	  sprintf(tmp,"..%d",curr);
	else
	  sprintf(tmp,",%d",curr);
	strcat(s,tmp);
      }
      if ( *border )
	strcat(s,",");
    }
  }
  strcat(s,">");
  return s;
}

char *make_indent (int num_tabs)
{
  char *new;
  if (num_tabs <= 0)
    return "";
  new = (char *)Malloc ((2 * num_tabs + 1) * sizeof (char));
  num_tabs *= 2;
  new[num_tabs] = '\0';
  while (--num_tabs >= 0) {
    new[num_tabs] = ' ';
  }
  return new;
}

#ifndef OPTIMIZE

void print_type_modifier (char *indent, char *string, TypeData *data)
{
  if (data) {
    printf ("%s%s: type modifier:\n", indent, string);
    printf ("%s%s  : size: %d\n", indent, string, TYPEDATA_SIZE(data));
    printf ("%s%s  : memsize: %d\n", indent, string, TYPEDATA_MEMSIZE(data));
    printf ("%s%s  : alignment: %d\n", indent, string, TYPEDATA_ALIGN(data));
    if (TYPEDATA_BORDER(data))
      printf ("%s%s  : byte order: %s\n", indent, string,
	       byteorder_to_string(TYPEDATA_BORDER(data)));
    else
      printf ("%s%s  : byte order: NULL\n", indent, string);
    if (TYPEDATA_OFFSET(data) != DEFAULT_OFFSET)
      printf ("%s%s  : offset: %d\n", indent, string,
	      TYPEDATA_OFFSET(data));
    else
      printf ("%s%s  : offset: INT_MAX = default\n", indent, string);
    printf ("%s%s  : complete: %d\n", indent, string, TYPEDATA_COMPLETE(data));
  }
}

static const char *op_to_string (int op)
{
  char *s;
  switch (op) {
  case DOT:
    s = "DOT";
    break;
    
  case ARRAY:
    s = "ARRAY";
    break;
    
  case PLUS:
    s = "PLUS";
    break;
    
  case MINUS:
    s = "MINUS";
    break;
    
  case STAR:
    s = "STAR";
    break;
    
  case POINTER:
    s = "POINTER";
    break;

  case ADDROF:
    s = "ADDROF";
    break;
  
  case RETURN:
    s = "RETURN";
    break;
  
  case SIZEOF:
    s = "SIZEOF";
    break;

  case DIV:
    s = "DIV";
    break;
    
  case MOD:
    s = "MOD";
    break;
    
  case AND:
    s = "AND";
    break;
    
  case PLUSPLUS:
    s = "PLUSPLUS";
    break;
    
  case MINMIN:
    s = "MINMIN";
    break;
    
  case LSHIFT:
    s = "LSHIFT";
    break;
    
  case RSHIFT:
    s = "RSHIFT";
    break;
    
  case GE:
    s = "GE";
    break;
    
  case LE:
    s = "LE";
    break;
    
  case NEQ:
    s = "NEQ";
    break;
    
  case EQEQ:
    s = "EQEQ";
    break;
    
  case TILDE:
    s = "TILDE";
    break;
    
  case BANG:
    s = "BANG";
    break;
    
  case LT:
    s = "LT";
    break;
    
  case GT:
    s = "GT";
    break;
    
  case COMMA:
    s = "COMMA";
    break;
    
  case EQUAL:
    s = "EQUAL";
    break;
    
  default:
    s = "unknown";
    break;
  }
  return s;
}

void print_operator (int level, int num_tabs, Expr *theexpr)
{
  char *indent = make_indent(num_tabs);
  printf ("%sOperator: %s\n", indent, op_to_string (EXPR_OP_TYPE(theexpr)));
  if (EXPR_OP_LEFT(theexpr)) {
    printf ("%sLeft: \n", indent);
    print_expr(level, num_tabs + 1, EXPR_OP_LEFT(theexpr));
  }
  else {
    printf ("%sLeft: NULL\n", indent);
  }
  if (EXPR_OP_RIGHT(theexpr)) {
    printf ("%sRight: \n", indent);
    print_expr(level, num_tabs + 1, EXPR_OP_RIGHT(theexpr));
  }
  else {
    printf ("%sRight: NULL\n", indent);
  }
}

void print_expr (int level, int num_tabs, Expr *theexpr)
{
  char *indent;
  char *units;

  if ( theexpr == NULL)
    return;
  else if (level <= Debug_Level) {
    indent = make_indent(num_tabs);
    if (!theexpr) {
      printf ("%sNULL expression pointer\n", indent);
      return;
    }
    switch (EXPR_TYPE(theexpr)) {
    case CONST_EXPR:
      if (EXPR_CONST_UNIT (theexpr) == BIT_UNIT)
        units = " in bits";
      else if (EXPR_CONST_UNIT (theexpr) == BYTE_UNIT)
        units = " in bytes";
      else
        units = "";
      printf ("%sConstant expr%s: ", indent, units);      
      print_constant (level, EXPR_CONST(theexpr));
      break;
      
    case OP_EXPR:
      printf ("%sOp expr:\n", indent);
      print_operator (level, num_tabs + 1, theexpr);
      break;
      
    case SYMINFO_EXPR:
      printf ("%sSymInfo expr:\n", indent);
      printf ("%sNum: %d\n", indent, SYMINFO_NUM (EXPR_SYMINFO (theexpr)));
      printf ("%sSym:\n", indent);
      print_sym (level, num_tabs + 1, "", EXPR_SYM(theexpr));
      printf ("%sTypedata:\n", indent);
      print_type_modifier (indent, "", EXPR_TYPEDATA(theexpr));
      printf ("%sBase Sym:\n", indent);
      print_sym (level, num_tabs + 1, "", EXPR_BASESYM(theexpr));
      printf ("%sOffset:\n", indent);
      print_expr (level, num_tabs + 1, EXPR_OFFSET (theexpr));
      break;
      
    case IDENT_EXPR:
      printf ("%sIdent expr: %s\n", indent, EXPR_IDENT(theexpr));
      break;
      
    default:
      printf ("%sUnknown expr type\n", indent);
      break;
    }
    printf ("\n");
  }
}

void print_structdef (Type **t, FILE *out)
{
  dprintf (900, "Entering print_structdef: %x", *t);
  print_type_list (0, 0, "", *t);
  dprintf (900, "Exiting print_structdef");
}

void ptabprint_sym (Symbol **s, char *string)
{
  print_sym (0, 0, string, *s);
}

void _print_struct_tab (Hash_Tab *table)
{
  ptab (table, print_structdef, stdout, 0);
}

void print_union_table (int level)
{
  if (level <= Debug_Level) {
    printf ("--------------------------------\n");
    printf ("Union Table:\n");
    _print_struct_tab (Union_Defs);
    printf ("--------------------------------\n");
  }
}

void print_struct_table (int level)
{
  if (level <= Debug_Level) {
    printf ("--------------------------------\n");
    printf ("Struct Table:\n");
    _print_struct_tab (Struct_Defs);
    printf ("--------------------------------\n");
  }
}

void _print_sym_tab (Hash_Tab *table)
{
  ptab (table, ptabprint_sym, "", 0);
}

void print_type_table (int level)
{
  if (level <= Debug_Level) {
    printf ("--------------------------------\n");
    printf ("Type Table:\n");
    _print_sym_tab (Type_Tab);
    printf ("--------------------------------\n");
  }
}

void print_sym_table (int level)
{
  if (level <= Debug_Level) {
    printf ("--------------------------------\n");
    printf ("Symbol Table:\n");
    _print_sym_tab (Symbol_Tab);
    printf ("--------------------------------\n");
  }
}

void ptabprint_syminfo (SymInfo **info, char *string)
{
  printf ("Num: %d\n", SYMINFO_NUM (*info));
  print_sym (0, 0, string, SYMINFO_BASESYM (*info));
}

void print_constant (int level, Constant *theconst)
{
  if (level <= Debug_Level) {
    switch (CONST_TYPE(theconst)) {
    case INT_CONST:
      printf ("integer: %ld\n", CONST_INT(theconst));
      break;
      
    case STRING_CONST:
      printf ("string: %s\n", CONST_STRING(theconst));
      break;
      
    case DOUBLE_CONST:
      printf ("double: %f\n", CONST_DOUBLE(theconst));
      break;
      
    default:
      printf ("Bad constant\n");
      break;
    }
  }
}

void print_type_spec_type (Type* type)
{
  printf ("%s", spec_type_to_string (type));
  printf ("\n");
}

void print_type (int level, int num_tabs, char *string, Type *type)
{
  char *indent;
  int i;

  dprintf (1000, "Entering print_type");
  if (level <= Debug_Level && type) {
    indent = make_indent(num_tabs);
    printf ("%s%s: type class: %s\n", indent, string,
	    TYPE_CLASS(type) == SPECIFIER ? "specifier" : "Declarator");
    printf ("%s%s: type tdef: %d\n", indent, string, TYPE_TDEF(type));
    printf ("%s%s: type native: %d\n", indent, string, TYPE_NATIVE(type));
    printf ("%s%s: type addnative: %d\n", indent, string, 
            TYPE_ADDNATIVE(type));
    if (TYPE_CLASS(type) == SPECIFIER) {
      printf ("%s%s: spec type: ", indent, 
	      string);
      print_type_spec_type(type);
      print_type_modifier (indent, string, TYPE_MODIFIERS(type));
      if (TYPE_SPEC_TYPE(type) & (UNION_TYPE | STRUCT_TYPE)) {
	if (TYPE_SPEC_STRUCTDEF(type)) {
	  if ( TYPE_STRUCTDEF_NAME(type) )
	    printf ("%s%s: name: <%s>\n", indent, string, 
		    TYPE_STRUCTDEF_NAME(type));
	  else
	    printf ("%s%s: no name\n", indent, string );
	  printf ("%s%s: level: %d\n", indent, string,
		  TYPE_STRUCTDEF_LEVEL(type));
	  printf ("%s%s: fields:\n", indent, string);
	  print_sym_list (level, num_tabs + 1, 
			  string, TYPE_STRUCTDEF_FIELDS(type));
	}
	else {
	  printf ("%s%s: no structdef\n", indent, string);
	}
      }
    }
    else {
      switch (TYPE_DECL_TYPE(type)) {
      case BITFIELD_TYPE:
	printf ("%s%s: declarator type: bitfield\n", indent, string);
	printf ("%s%s: num_ele: %d\n", indent, string, TYPE_DECL_NUMEL(type));
        if (TYPE_DECL_BITINFO (type)) {
          printf ("%s%s: bits:", indent, string);
          for (i = 0; i < MAX_BITFIELD_SIZE; i++)
            printf (" %2d", TYPE_DECL_BITINFO(type)[i]);
          printf ("\n");
	  printf ("%s%s: bits: %s\n", indent, string, 
		  bitorder_to_string (TYPE_DECL_BITINFO (type)));
        }
        else {
          printf ("%s%s: no bit info\n", indent, string);
        }
        print_type_modifier (indent, string, TYPE_MODIFIERS(type));
	break;
	
      case ARRAY_TYPE:
	printf ("%s%s: declarator type: array\n", indent, string);
	printf ("%s%s: num_ele: %d\n", indent, string, TYPE_DECL_NUMEL(type));
	print_type_modifier (indent, string, TYPE_MODIFIERS(type));
	break;
	
      case FUNCTION_TYPE:
	printf ("%s%s: declarator type: function\n", indent, string);
	print_type_modifier (indent, string, TYPE_MODIFIERS(type));
	printf ("%s%s: parameters: %d\n", indent, string,TYPE_DECL_NUMEL(type));
	print_sym_list (level, num_tabs + 1, string, TYPE_DECL_PARAMS(type));
	break;
	
      case MODIFIER_TYPE:
	printf ("%s%s: declarator type: modifier type\n", indent, string);
	print_type_modifier (indent, string, TYPE_DECL_TYPEDATA(type));
	break;
	
      case POINTER_TYPE:
	printf ("%s%s: declarator type: pointer\n", indent, string);
	print_type_modifier (indent, string, TYPE_DECL_TYPEDATA(type));
	break;
	
      default:
	printf ("%s%s: type: unknown declarator\n", indent, string);
	break;
      }
    }
    printf ("\n");
  }
  dprintf (1000, "Exiting print_type");
}

void print_type_list (int level, int num_tabs, char *string, Type *type)
{
  dprintf (1000, "Entering print_type_list");
  if (level <= Debug_Level) { 
    while (type) {
      print_type (level, num_tabs, string, type);
      type = type->next;
    }
  }
  dprintf (1000, "Exiting print_type_list");
}

void print_sym (int level, int num_tabs, char *string, Symbol *sym)
{
  char *indent;

  dprintf (1000, "Entering print_sym");

  if (level <= Debug_Level && sym) {
    indent = make_indent(num_tabs);
    printf ("%s%s: symbol name: %s\n", indent, string, SYM_NAME(sym));
    printf ("%s%s: symbol level: %d\n", indent, string, SYM_LEVEL(sym));
    printf ("%s%s: symbol local_name: %s\n", indent, string,
            SYM_LOCAL_NAME(sym));
    if (SYM_TYPEDATA(sym)) {
      printf ("%s%s: size: %d\n", indent, string, SYM_SIZE(sym));
      printf ("%s%s: memsize: %d\n", indent, string, SYM_MEMSIZE(sym));
      printf ("%s%s: alignment: %d\n", indent, string, SYM_ALIGN(sym));
      if (SYM_BORDER(sym))
	printf ("%s%s: byte order: %s\n", indent, string,
		 byteorder_to_string(SYM_BORDER(sym)));
      else
	printf ("%s%s: byte order: NULL\n", indent, string);
      if (SYM_OFFSET(sym) != DEFAULT_OFFSET)
	printf ("%s%s: offset: %d\n", indent, string, SYM_OFFSET(sym));
      else
	printf ("%s%s: offset: INT_MAX = default\n", indent, string);
      printf ("%s%s: complete: %d\n", indent, string, SYM_COMPLETE(sym));
    }
    else {
      printf ("%s%s: no typedata\n", indent, string);
    }
    printf ("%s%s: symbol type: \n", indent, string);
    print_type_list (level, num_tabs + 1, string, SYM_TYPE(sym));
    printf ("\n");
  }
  dprintf (1000, "Exiting print_sym");
}

void print_sym_list (int level, int num_tabs, char *string, Symbol *sym)
{
  dprintf (1000, "Entering print_sym_list");
  if (level <= Debug_Level) {
    while (sym) {
      print_sym (level, num_tabs, string, sym);
      sym = sym->next;
    }
  }
  dprintf (1000, "Exiting print_sym_list");
}

void dprintf(int level, char *fmt, ... )
{
  va_list args;

  va_start(args,fmt);
  if (level <= Debug_Level) {
    fprintf (stdout, "debug: ");
    vfprintf (stdout, fmt, args);
    fprintf (stdout, "\n");
  }
  va_end (args);
}


#endif /* OPTIMIZE */





