%{
/*
 * $RCSfile: grammar.y,v $
 *
 * Universal Stub Compiler
 *
 * Copyright (c) 1994    Arizona Board of Regents
 *
 * $Log: grammar.y,v $
 * Revision 1.1  1994/10/13 17:46:05  dorgival
 * Initial revision
 *
 * Revision 1.1  1994/10/13  17:43:30  dorgival
 * Initial revision
 *
 */

/*
 * Grammar definition: basically ANSI C with extensions for type annotation
 */

#include "usc.h"
#include "symbol.h"
#include "debug.h"
#include "buffer.h"
#include "codegen_export.h"
#include "expr.h"
#include "preprocess.h"
#ifdef NO_ALLOCA
#include <stdlib.h>
#else
#include <alloca.h>
#endif

extern int yylex();
extern int yyerror(char *s);

extern void check_config(void);
%}

%union {
   Constant theconst[1];
   Constant *theconstptr;
   Symbol *thesym;
   Type *thetype;
   Expr *theexpr;
   int theop;
}

%token <theconst> IDENT INT LONG CHAR SHORT FLOAT DOUBLE
%token <theconst> SIGNED UNSIGNED EXTERN
%token <theconst> AUTO CONST REGISTER
%token <theconst> STRUCT UNION TYPEDEF
%token <theconst> ICON FCON STRING
%token <theconst> INLINE MACRO

%token <thetype>  TYPEDEFNAME

%token <theop>    ARROW DOT PLUSPLUS MINMIN LSHIFT RSHIFT
%token <theop>    ARRAY POINTER SIZEOF ADDROF
%token <theop>    LE GE GT LT EQEQ NEQ ANDAND OROR
%token <theop>    '+' '-' '*' '/' '%'
%token <theop>    PLUS MINUS STAR DIV MOD
%token <theop>    AND OR XOR
%token <theop>    '~' '!' ','
%token <theop>    TILDE BANG COMMA
%token <theop>    EQUAL RETURN

%token VOID STATIC ENUM VOLATILE
%token HASH PRAGMA DOTDOT DOTDOTDOT
%token CASE DEFAULT IF ELSE SWITCH WHILE DO FOR GOTO CONTINUE BREAK
%token MULEQ DIVEQ MODEQ ADDEQ SUBEQ LSHIFTEQ RSHIFTEQ ANDEQ XOREQ OREQ

%type <thesym> translation_unit
%type <thesym> external_definition
%type <thesym> function_or_declaration
%type <thesym> function_definition
%type <thesym> declaration declaration_list
%type <thesym> declarator declarator2
%type <thesym> init_declarator init_declarator_list
%type <thesym> parameter_list parameter_type_list parameter_declaration
%type <thesym> struct_declarator struct_declarator_list
%type <thesym> struct_declaration_list struct_declaration

%type <thesym> struct_field_declarator struct_field_declarator2
%type <thetype> struct_type_info_list

%type <thetype> type_specifier type_specifier_list
%type <thetype> type_name
%type <thetype> declaration_specifiers
%type <thetype> struct_or_union struct_or_union_specifier
%type <thetype> type_info_list
%type <thetype> bitfield_constructor bitfield_description
%type <thetype> bit_order bits_list

%type <theop>   structop incop unary_operator shiftop relop equop
%type <theop>   pointer assignment_operator
%type <theop>   code_type_specifier

%type <theexpr> function_body compound_statement
%type <theexpr> jump_statement selection_statement iteration_statement
%type <theexpr> expr expression_statement statement_list statement
%type <theexpr> labeled_statement
%type <theexpr> byte_order

%type <theconstptr> bit_range byte_list byte_range

%type <theexpr> primary_expr postfix_expr argument_expr_list unary_expr
%type <theexpr> cast_expr multiplicative_expr additive_expr shift_expr
%type <theexpr> relational_expr equality_expr and_expr exclusive_or_expr
%type <theexpr> inclusive_or_expr logical_and_expr logical_or_expr
%type <theexpr> conditional_expr assignment_expr constant_expr

%type <theconst> simple_type pragma_type  storage_class_specifier

%start translation_unit
%%

primary_expr
    : IDENT        { $$ = ident_to_expr ($1); print_expr (1000, 0, $$); }
    | ICON         { $$ = const_to_expr ($1); print_expr (1000, 0, $$); }
    | FCON         { $$ = const_to_expr ($1); print_expr (1000, 0, $$); }
    | STRING       { $$ = const_to_expr ($1); print_expr (1000, 0, $$); }
    | '(' expr ')' { $$ = $2; }
    ;

structop
    : ARROW | DOT;

incop
    : PLUSPLUS | MINMIN ;

postfix_expr
    : primary_expr
    | postfix_expr '[' expr ']'   { $$ = do_expr_op($1, ARRAY, $3); }
    | postfix_expr structop IDENT { $$ = do_expr_op($1, $2, ident_to_expr($3));}
    | postfix_expr incop          { $$ = postfix_op ($2, $1); }
    | postfix_expr '(' ')' {
      $$ = NULL;
      error("procedure call not allowed");
    }
    | postfix_expr '(' argument_expr_list ')' {
      $$ = NULL;
      error("procedure call not allowed");
    }
    ;

argument_expr_list
    : assignment_expr
    | argument_expr_list ',' assignment_expr
    ;

unary_expr
    : postfix_expr
    | incop unary_expr         { $$ = prefix_op ($1, $2); }
    | unary_operator cast_expr { $$ = prefix_op ($1, $2); }
    | SIZEOF '(' type_name ')' {
      $$ = sizeof_type( $3 );
    }
    | SIZEOF unary_expr        {
      $$ = prefix_op( $1, $2 );
    }
    ;

unary_operator
    : AND { $$ = ADDROF; }
    | '*' { $$ = POINTER; }
    | '+'
    | '-'
    | '~'
    | '!'
    ;

cast_expr
    : unary_expr
    | '(' type_name ')' cast_expr {
      $$ = $4;
      error("typecasting ignored");
    }
    ;

multiplicative_expr
    : cast_expr
    | multiplicative_expr '*' cast_expr { $$ = do_expr_op($1, $2, $3); }
    | multiplicative_expr '/' cast_expr { $$ = do_expr_op($1, $2, $3); }
    | multiplicative_expr '%' cast_expr { $$ = do_expr_op($1, $2, $3); }
    ;

additive_expr
    : multiplicative_expr
    | additive_expr '+' multiplicative_expr { $$ = do_expr_op($1, $2, $3); }
    | additive_expr '-' multiplicative_expr { $$ = do_expr_op($1, $2, $3); }
    ;

shiftop
    : LSHIFT | RSHIFT ;

shift_expr
    : additive_expr
    | shift_expr shiftop additive_expr { $$ = do_expr_op($1, $2, $3); }
    ;

relop   : LE | GE | LT | GT

relational_expr
    : shift_expr
    | relational_expr relop shift_expr { $$ = do_expr_op($1, $2, $3); }
    ;

equop   : EQEQ | NEQ ;

equality_expr
    : relational_expr
    | equality_expr equop relational_expr { $$ = do_expr_op($1, $2, $3); }
    ;

and_expr
    : equality_expr
    | and_expr AND equality_expr { $$ = do_expr_op($1, $2, $3); }
    ;

exclusive_or_expr
    : and_expr
    | exclusive_or_expr XOR and_expr { $$ = do_expr_op($1, $2, $3); }
    ;

inclusive_or_expr
    : exclusive_or_expr
    | inclusive_or_expr OR exclusive_or_expr { $$ = do_expr_op($1, $2, $3); }
    ;

logical_and_expr
    : inclusive_or_expr
    | logical_and_expr ANDAND inclusive_or_expr { $$ = do_expr_op($1, $2, $3); }
    ;

logical_or_expr
    : logical_and_expr
    | logical_or_expr OROR logical_and_expr { $$ = do_expr_op($1, $2, $3); }
    ;

conditional_expr
    : logical_or_expr
    | logical_or_expr '?' logical_or_expr ':' conditional_expr {
      $$ = NULL;
      error("'?' operator not implemented yet");
    }
    ;

assignment_expr
    : conditional_expr
    | unary_expr assignment_operator assignment_expr {
      $$ = do_expr_op ($1, $2, $3);
    }
    ;

assignment_operator
    : EQUAL
    | ADDEQ     { error("'+=' not allowed"); }
    | SUBEQ     { error("'-=' not allowed"); }
    | MULEQ     { error("'*=' not allowed"); }
    | DIVEQ     { error("'/=' not allowed"); }
    | MODEQ     { error("'%=' not allowed"); }
    | ANDEQ     { error("'&=' not allowed"); }
    | XOREQ     { error("'|=' not allowed"); }
    | OREQ      { error("'^=' not allowed"); }
    | LSHIFTEQ  { error("'<<=' not allowed"); }
    | RSHIFTEQ  { error("'>>=' not allowed"); }
    ;

expr
    : assignment_expr { $$ = $1; print_expr (700, 0, $$); }
    | expr ',' assignment_expr {
      $$ = do_expr_op ($1, $2, $3);
      print_expr (700, 0, $$);
    }
    ;

constant_expr
    : conditional_expr { print_expr (1000, 0, $1); }
    ;

declaration
    : declaration_specifiers ';'  { $$ = NULL; }
    | declaration_specifiers init_declarator_list ';' {
      Symbol *sym;
      sym = distribute_specifier ($1, $2);
      if ( !do_preprocess ) 
	set_all_sym_type_info (sym);
      $$ = add_to_sym_or_type_table (sym);
      if ( !do_preprocess && TYPE_TDEF(SYM_TYPE(sym)) && preserve_type_names )
	write_fake_type( HeaderFile, sym );
    }
    ;

declaration_specifiers
    : storage_class_specifier {
      $$ = NULL;
      error("'typedef' must be followed by type definition");
    }
    | storage_class_specifier declaration_specifiers {
      /* must confirm storage_class_specifier is typedef? */
      $$ = make_typedef ($2);
    }
    | type_specifier
    | type_specifier declaration_specifiers {
      print_type_table(900);
      $$ = or_specifier ($1, $2);
      print_type_table(900);
    }
    ;

init_declarator_list
    : init_declarator
    | init_declarator_list ',' init_declarator { $$ = cat_sym( $1, $3 ); }
    ;

init_declarator
    : declarator
    | declarator EQUAL initializer { 
      $$ = $1;
      error("no variable initialization allowed");
    }
    ;

storage_class_specifier
    : TYPEDEF
    | EXTERN   { error("'extern' storage class not allowed"); }
    | AUTO     { error("'auto' storage class not allowed"); }
    | REGISTER { error("'register' storage class not allowed"); }
    ;

simple_type
    : CHAR
    | SHORT
    | INT
    | LONG
    | SIGNED
    | UNSIGNED
    ;


type_specifier
    : simple_type {
      $$ = make_specifier ($1);
      print_type (800, 0, "type_specifier", $$);
    }
    | FLOAT       { $$=NULL; error("'float' type not allowed"); }
    | DOUBLE      { $$=NULL; error("'double' type not allowed"); }
    | CONST       { $$=NULL; error("'const' types not recognized"); }
    | VOLATILE    { $$=NULL; error("'volatile' type not recognized"); }
    | VOID        { $$=NULL; }
    | struct_or_union_specifier
    | enum_specifier { $$=NULL; }
    | TYPEDEFNAME
    ;

struct_or_union_specifier
    : struct_or_union IDENT '{' struct_declaration_list'}' {
      Type *temp;
      print_type (800, 0, "struct_or_union", $1);
      temp = find_struct_def (CONST_STRING($2));
      if (temp)
	crash ("Struct <%s> already defined", CONST_STRING($2));
      else {
	temp = finish_struct_def ($1, CONST_STRING($2), $4);
	add_to_struct_defs (temp, NestLevel);
	$$ = temp;
      }
    }
    | struct_or_union '{' struct_declaration_list '}' {
      Type *temp;
      temp = finish_struct_def ($1, NULL, $3);
      $$ = temp;
    }
    | struct_or_union IDENT {
      $$ = find_struct_def (CONST_STRING($2));
      if ($$ == NULL) crash ("Undefined struct: %s", CONST_STRING($2));
    }
    ;

struct_or_union
    : STRUCT { $$ = make_specifier ($1); }
    | UNION  { $$ = make_specifier ($1); }
    ;

struct_declaration_list
    : struct_declaration
    | struct_declaration_list struct_declaration {
      $$ = cat_sym ($1, $2);
      print_sym_list (800, 0,"struct_decltn_list", $$);
    }
    ;

struct_declaration
    : type_specifier_list struct_declarator_list ';' {
      $$ = distribute_specifier ($1, $2);
      print_sym_list(800, 0,"struct_declaration", $$);
    }
    ;

struct_declarator_list
    : struct_declarator
    | struct_declarator_list ',' struct_declarator {
      $$ = cat_sym ($1, $3);
      print_sym_list (800, 0, "struct_declr_list", $$);
    }
    ;

struct_declarator
    : struct_field_declarator { print_sym(800,0,"struct_declr", $1); }
    | ':' bitfield_description {
      $$ = NULL;
      error("not named bitfields not allowed");
    }
    | struct_field_declarator ':' bitfield_description {
      $$=make_bitfield($1,$3);
    }
    ;

/* ---------- */
struct_field_declarator
    : struct_field_declarator2       
    | pointer struct_field_declarator2 {
      error("no pointers allowed as struct fields");
      $$ = NULL;
      /* must insert pointer info */
    }
    ;

struct_field_declarator2
    : IDENT                                           {
      $$ = ident_to_sym(CONST_STRING($1));
    }
    | '(' struct_field_declarator ')'                 { $$ = $2; }
    | struct_field_declarator2 '[' ']'                {
      error( "array definition misses length" );
    }
    | struct_field_declarator2 '[' constant_expr ']'  {
      $$ = make_array ($1, $3);
    }
    | struct_field_declarator2 '(' struct_type_info_list ')' {
      $$ = add_type_info ($1, $3);
    }
    | struct_field_declarator2 '(' ')' {
      $$ = NULL;
      error("functions are not allowed as fields");
    }
    | struct_field_declarator2 '(' parameter_type_list ')' {
      $$ = NULL;
      error("functions are not allowed as fields");
      /* must combine parameter type info */
    }
    | struct_field_declarator2 '(' parameter_identifier_list ')' {
      $$ = NULL;
      error("functions are not allowed as fields");
    }
    ;

struct_type_info_list
    : ICON ',' ICON ',' ICON ',' byte_order {
      $$ = make_type_info ($1, $3, NULL, $5, $7 );
    }
    ;

/* --------- */

bitfield_description
    : ICON /* constant_expr */ {
      if ( !do_preprocess )
	error("all bitfields must be annotated");
      $$ = make_native_bitfield (const_to_expr($1));
    }
    | ICON /* constant_expr */ '(' bitfield_constructor ')' {
      $$ = make_native_bitfield (const_to_expr($1));
      $$ = add_bitfield_constructor ($$, $3);
    }
    ;

bitfield_constructor
    : ICON ',' ICON ',' ICON ',' bit_order {
      $$ = make_bitfield_constructor( $1, $3, $5, $7 );
    }
    ;

bit_order
    : LT bits_list GT { $$ = $2; }
    ;

bits_list
    : bits_list ',' bit_range { $$ = add_more_bits ($1, $3); }
    | bit_range               { $$ = add_more_bits (NULL, $1); }
    ;

bit_range
    : ICON             { $$ = EXPR_CONST ( const_to_expr ($1) ); }
    | ICON DOTDOT ICON { $$ = cat_const ( EXPR_CONST ( const_to_expr ($1) ), EXPR_CONST ( const_to_expr ($3 ) ) ); }
    ;

enum_specifier
    : ENUM '{' enumerator_list '}'
    | ENUM IDENT '{' enumerator_list '}'
    | ENUM IDENT
    ;

enumerator_list
    : enumerator
    | enumerator_list ',' enumerator
    ;

enumerator
    : IDENT
    | IDENT EQUAL constant_expr
    ;

declarator
    : declarator2 {
      $$ = $1;
    }
    | pointer declarator2 {
      if ( $1 == STAR ) {
	$$ = make_pointer( $2 );
      } else {
	$$ = $2;
      }
    }
    ;

declarator2
    : IDENT                              { $$ = ident_to_sym(CONST_STRING($1));}
    | '(' declarator ')'                 { $$ = $2; }
    | declarator2 '[' ']'                {
      error( "array definition misses length" );
    }
    | declarator2 '[' constant_expr ']'  { $$ = make_array ($1, $3); }
    | declarator2 '(' type_info_list ')' { $$ = add_type_info ($1, $3); }
    | declarator2 '(' ')' {
      $$ = make_function( $1, NULL );
    }
    | declarator2 '(' parameter_type_list ')' {
      $$ = make_function( $1, $3 );
    }
    | declarator2 '(' parameter_identifier_list ')' {
      $$ = NULL;
      error("parameters in parameter list must be typed");
    }
    ;

type_info_list
    : ICON ',' ICON ',' ICON ',' byte_order {
      $$ = make_type_info ($1, $3, $5, NULL, $7 );
    }
    ;

byte_order
    : LT byte_list GT { $$ = const_to_expr( $2 ); }
    | ICON { $$ = NULL; } /* should check if constant is zero? */
    ;

byte_list
    : byte_range                      { $$ = add_more_bytes(NULL,$1); }
    | byte_list ',' byte_range        { $$ = add_more_bytes(  $1,$3); }
    ;

byte_range
    : ICON             { $$ = EXPR_CONST ( const_to_expr ($1) ); }
    | ICON DOTDOT ICON { $$ = cat_const( EXPR_CONST ( const_to_expr ($1)), EXPR_CONST ( const_to_expr ($3) ) ); }
    ;


pointer
    : '*'
    | '*' type_specifier_list           {
      $$ = ' ';
      error("only a simple '*' is allowed");
    }
    | '*' pointer {
      $$ = ' ';
      error("only a simple '*' is allowed");
    }
    | '*' type_specifier_list pointer {
      $$ = ' ';
      error("only a simple '*' is allowed");
    }
    ;

type_specifier_list
    : type_specifier
    | type_specifier type_specifier_list { $$ = or_specifier($1, $2); }
    ;

parameter_identifier_list
    : identifier_list
    | identifier_list ',' DOTDOTDOT {
      error("stubs must have fixed parameter lists");
    }
    ;

identifier_list
    : IDENT
    | identifier_list ',' IDENT
    ;

parameter_type_list
    : parameter_list
    | parameter_list ',' DOTDOTDOT {
      $$ = $1;
      error("stubs must have fixed parameter lists");
    }
    ;

parameter_list
    : parameter_declaration {
      $$ = $1;
      /*
      if ( $1 != NULL ) {
	printf( "Adding %s\n", SYM_NAME($1) );
	add_param(1, $1, NORMAL_PARAM);
      }
      */
    }
    | parameter_list ',' parameter_declaration {
      if ( $1 != NULL )
	$$ = cat_sym( $1, $3 );
      else
	$$ = $3;
      /*
      if ( $3 != NULL ) {
	printf( "Adding %s\n", SYM_NAME($3) );
	add_param(1, $3, NORMAL_PARAM);
      }
      */
    }
    ;

parameter_declaration
    : type_specifier_list declarator {
      Symbol *temp;
      temp = distribute_specifier ($1, $2);
      if ( !do_preprocess )  
	set_all_sym_type_info (temp);
      NestLevel++;
      $$ = add_to_sym_or_type_table (temp);

      if ( SYM_IS_STRUCT_OR_UNION(temp) ) {
	error("structs and unions cannot be passed by value to stubs");
      }
      else if ( !SYM_IS_ARRAY(temp) && !SYM_IS_POINTER(temp) ) {
	dont_deref( SYM_LOCAL_NAME(temp) );
      }

      NestLevel--;
    }
    | type_name {
      $$ = NULL;
      error("parameters must have names associated with them");
    }
    ;

type_name
    : type_specifier_list
    | type_specifier_list abstract_declarator {
      error("only symple USC type allowed as 'type name'");
      $$ = NULL;
    }
    ;

abstract_declarator
    : pointer
    | abstract_declarator2
    | pointer abstract_declarator2
    ;

abstract_declarator2
    : '(' abstract_declarator ')'
    | '[' ']'
    | '[' constant_expr ']'
    | abstract_declarator2 '[' ']'
    | abstract_declarator2 '[' constant_expr ']'
    | '(' ')'
    | '(' parameter_type_list ')'
    | abstract_declarator2 '(' ')'
    | abstract_declarator2 '(' parameter_type_list ')'
    ;

initializer
    : assignment_expr
    | '{' initializer_list '}'
    | '{' initializer_list ',' '}'
    ;

initializer_list
    : initializer
    | initializer_list ',' initializer
    ;

statement
    : labeled_statement
    | compound_statement
    | expression_statement
    | selection_statement
    | iteration_statement
    | jump_statement
    ;

labeled_statement
    : IDENT ':' statement {
      $$ = $3;
    }
    | CASE constant_expr ':' statement {
      $$ = NULL; 
      error("no case statements allowed");
    }
    | DEFAULT ':' statement {
      $$ = NULL; 
      error("no default statements allowed");
    }
    ;

compound_statement
    : '{' '}'                                  { $$ = NULL; }
    | '{' statement_list '}'                   { $$ = $2; }
    | '{' declaration_list '}'                 {
      $$ = NULL;
      error("local definitions not allowed");
    }
    | '{' declaration_list statement_list '}'  {
      $$ = NULL;
      error("local definitions not allowed");
    }
    ;

declaration_list
    : declaration
    | declaration_list declaration {
      $$ = cat_sym ($1, $2);
      print_sym_list (800, 0,"struct_decltn_list", $$);
    }
    ;

statement_list
    : statement
    | statement statement_list { $$ = cat_expr( $1, $2); }
    ;

expression_statement
    : ';'        { $$ = NULL; }
    | expr ';'   { if (!do_preprocess) $$ = clean_expr( $1 ); }
    ;

selection_statement
    : IF '(' expr ')' statement {
      $$ = NULL;
      error("'if' not allowed");
    }
    | IF '(' expr ')' statement ELSE statement {
      $$ = NULL;
      error("'if-else' not allowed");
    }
    | SWITCH '(' expr ')' statement {
      $$ = NULL;
      error("'switch' not allowed");
    }
    ;

iteration_statement
    : WHILE '(' expr ')' statement {
      $$ = NULL;
      error("'while' not allowed");
    }
    | DO statement WHILE '(' expr ')' ';' {
      $$ = NULL;
      error("'do-while' not allowed");
    }
    | FOR '(' ';' ';' ')' statement {
      $$ = NULL;
      error("'for' not allowed");
    }
    | FOR '(' ';' ';' expr ')' statement {
      $$ = NULL;
      error("'for' not allowed");
    }
    | FOR '(' ';' expr ';' ')' statement {
      $$ = NULL;
      error("'for' not allowed");
    }
    | FOR '(' ';' expr ';' expr ')' statement {
      $$ = NULL;
      error("'for' not allowed");
    }
    | FOR '(' expr ';' ';' ')' statement {
      $$ = NULL;
      error("'for' not allowed");
    }
    | FOR '(' expr ';' ';' expr ')' statement {
      $$ = NULL;
      error("'for' not allowed");
    }
    | FOR '(' expr ';' expr ';' ')' statement {
      $$ = NULL;
      error("'for' not allowed");
    }
    | FOR '(' expr ';' expr ';' expr ')' statement {
      $$ = NULL;
      error("'for' not allowed");
    }
    ;

jump_statement
    : GOTO IDENT ';'  { $$=NULL; error("'goto' not allowed"); }
    | CONTINUE ';'    { $$=NULL; error("'continue' not allowed"); }
    | BREAK ';'       { $$=NULL; error("'break' not allowed"); }
    | RETURN ';'      { $$=make_op(NULL,$1,NULL); }
    | RETURN expr ';' { $$=prefix_op($1,$2); 
			if (!do_preprocess) $$=clean_expr($$); }
    ;

translation_unit
    : external_definition
    | translation_unit external_definition
    ;

external_definition
    : function_or_declaration { $$ = $1; 
				if (!configured && !do_preprocess) 
				  check_config(); 
			      }
    | preprocess_command      { $$ = NULL; }
    ;

code_type_specifier
    : STATIC                      { code_type |= GEN_STATIC; }
    | STATIC code_type_specifier  { code_type |= GEN_STATIC; }
    | INLINE                      {
				     if ( CODE_TYPE_SET )
				       error("code type conflict");
				     else
				       code_type |= GEN_INLINE;
				  }
    | INLINE code_type_specifier  {
				     if ( CODE_TYPE_SET )
				       error("code type conflict");
				     else
				       code_type |= GEN_INLINE;
				  }
    | MACRO                       {
				     if ( CODE_TYPE_SET )
				       error("code type conflict");
				     else
				       code_type |= GEN_MACROS;
				  }
    | MACRO code_type_specifier   {
				     if ( CODE_TYPE_SET )
				       error("code type conflict");
				     else
				       code_type |= GEN_MACROS;
				  }
    ;

function_or_declaration
    : function_definition   {
      Symbol* function = $1;
      if ( !CODE_TYPE_SET )
	code_type |= GEN_PLAIN;
      if ( !do_preprocess )
	make_stub( function );
      else
	print_source();
      $$ = NULL;
      code_type = DefaultCodeType;
    }
    | code_type_specifier function_definition {
      Symbol* function = $2;
      if ( !CODE_TYPE_SET )
	code_type |= GEN_PLAIN;
      if ( !do_preprocess )
	make_stub( function );
      else
	print_source();
      $$ = NULL;
      code_type = DefaultCodeType;
    }
    /*
    | INLINE function_definition   {
      Symbol* function = $2;
      code_type |= GEN_INLINE;
      if ( !do_preprocess )
	make_stub( function );
      else
	print_source();
      $$ = NULL;
      code_type = DefaultCodeType;
    }
    | MACRO function_definition   {
      Symbol* function = $2;
      code_type |= GEN_MACROS;
      if ( !do_preprocess )
	make_stub( function );
      else
	print_source();
      $$ = NULL;
      code_type = DefaultCodeType;
    }
    */
    | declaration           { 
      if (do_preprocess)
	dump_source();
      else
	print_source(); 
    }
    ;

preprocess_command
    : HASH PRAGMA pragma_type '(' type_info_list ')' {
      cpp_pragma_command (make_specifier($3), $5);
    }
    | HASH ICON STRING { cpp_line_command ($2, $3); }
    ;

pragma_type
    : INT
    | CHAR
    | SHORT
    | LONG
    ;

function_definition
    : declarator function_body {
      Constant default_return;
      set_to_string( &default_return, "int" );
      $$ = complete_function( make_specifier( &default_return ), $1, $2 );
    }
    | declaration_specifiers declarator function_body {
      $$ = complete_function( $1, $2, $3 );
    }
    ;

function_body
    : compound_statement
    | declaration_list compound_statement {
      error("no local variables allowed");
      $$ = NULL;
    }
    ;
%%
